"""This module provides a set of classes which underpin the data loading and
saving functionality provided by ``kedro.io``.
"""

from __future__ import annotations

import abc
import copy
import logging
import pprint
import sys
import warnings
from collections import namedtuple
from datetime import datetime, timezone
from functools import partial, wraps
from glob import iglob
from inspect import getcallargs
from operator import attrgetter
from pathlib import Path, PurePath, PurePosixPath
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    List,
    Literal,
    Protocol,
    TypeVar,
    runtime_checkable,
)

from cachetools import Cache, cachedmethod
from cachetools.keys import hashkey
from typing_extensions import Self

# These are re-exported for backward compatibility
from kedro.utils import (  # noqa: F401
    CLOUD_PROTOCOLS,
    HTTP_PROTOCOLS,
    _parse_filepath,
    load_obj,
)

if TYPE_CHECKING:
    import os
    from multiprocessing.managers import SyncManager


VERSION_FORMAT = "%Y-%m-%dT%H.%M.%S.%fZ"
VERSIONED_FLAG_KEY = "versioned"
VERSION_KEY = "version"
PROTOCOL_DELIMITER = "://"
TYPE_KEY = "type"

# Type alias for copy modes
TCopyMode = Literal["deepcopy", "copy", "assign"]


class DatasetError(Exception):
    """``DatasetError`` raised by ``AbstractDataset`` implementations
    in case of failure of input/output methods.

    ``AbstractDataset`` implementations should provide instructive
    information in case of failure.
    """

    pass


class DatasetNotFoundError(DatasetError):
    """``DatasetNotFoundError`` raised by ``DataCatalog``
    class in case of trying to use a non-existing dataset.
    """

    pass


class DatasetAlreadyExistsError(DatasetError):
    """``DatasetAlreadyExistsError`` raised by ``DataCatalog``
    class in case of trying to add a dataset which already exists in the ``DataCatalog``.
    """

    pass


class VersionNotFoundError(DatasetError):
    """``VersionNotFoundError`` raised by ``AbstractVersionedDataset`` implementations
    in case of no load versions available for the dataset.
    """

    pass


class VersionAlreadyExistsError(DatasetError):
    """``VersionAlreadyExistsError`` raised by ``DataCatalog``
    class when attempting to add a dataset to a catalog with a save version
    that conflicts with the save version already set for the catalog.
    """

    pass


_DI = TypeVar("_DI")
_DO = TypeVar("_DO")


class AbstractDataset(abc.ABC, Generic[_DI, _DO]):
    """``AbstractDataset`` is the base class for all dataset implementations.

    All dataset implementations should extend this abstract class
    and implement the methods marked as abstract.
    If a specific dataset implementation cannot be used in conjunction with
    the ``ParallelRunner``, such user-defined dataset should have the
    attribute `_SINGLE_PROCESS = True`.
    Example:
    ``` python
    from pathlib import Path, PurePosixPath
    import pandas as pd
    from kedro.io import AbstractDataset


    class MyOwnDataset(AbstractDataset[pd.DataFrame, pd.DataFrame]):
        def __init__(self, filepath, param1, param2=True):
            self._filepath = PurePosixPath(filepath)
            self._param1 = param1
            self._param2 = param2

        def load(self) -> pd.DataFrame:
            return pd.read_csv(self._filepath)

        def save(self, df: pd.DataFrame) -> None:
            df.to_csv(str(self._filepath))

        def _exists(self) -> bool:
            return Path(self._filepath.as_posix()).exists()

        def _describe(self):
            return dict(param1=self._param1, param2=self._param2)
    ```
    Example catalog.yml specification:
    ``` yaml
    my_dataset:
        type: <path-to-my-own-dataset>.MyOwnDataset
        filepath: data/01_raw/my_data.csv
        param1: <param1-value> # param1 is a required argument
        # param2 will be True by default
    ```
    """

    """
    Datasets are persistent by default. User-defined datasets that
    are not made to be persistent, such as instances of `MemoryDataset`,
    need to change the `_EPHEMERAL` attribute to 'True'.
    """
    _EPHEMERAL = False

    @classmethod
    def from_config(
        cls: type,
        name: str,
        config: dict[str, Any],
        load_version: str | None = None,
        save_version: str | None = None,
    ) -> AbstractDataset:
        """Create a dataset instance using the configuration provided.

        Args:
            name: Data set name.
            config: Data set config dictionary.
            load_version: Version string to be used for ``load`` operation if
                the dataset is versioned. Has no effect on the dataset
                if versioning was not enabled.
            save_version: Version string to be used for ``save`` operation if
                the dataset is versioned. Has no effect on the dataset
                if versioning was not enabled.

        Returns:
            An instance of an ``AbstractDataset`` subclass.

        Raises:
            DatasetError: When the function fails to create the dataset
                from its config.

        """
        try:
            class_obj, config = parse_dataset_definition(
                config, load_version, save_version
            )
        except Exception as exc:
            raise DatasetError(
                f"An exception occurred when parsing config "
                f"for dataset '{name}':\n{exc!s}"
            ) from exc

        try:
            dataset = class_obj(**config)
        except TypeError as err:
            raise DatasetError(
                f"\n{err}.\nDataset '{name}' must only contain arguments valid for the "
                f"constructor of '{class_obj.__module__}.{class_obj.__qualname__}'."
            ) from err
        except Exception as err:
            raise DatasetError(
                f"\n{err}.\nFailed to instantiate dataset '{name}' "
                f"of type '{class_obj.__module__}.{class_obj.__qualname__}'."
            ) from err
        return dataset

    def _init_config(self) -> dict[str, Any]:
        """Internal method to capture the dataset's initial configuration
        as provided during instantiation.

        This configuration reflects only the arguments supplied at `__init__` time
        and does not account for any runtime or dynamic changes to the dataset.

        Adds a key for the dataset's type using its module and class name and
        includes the initialization arguments.

        For `CachedDataset` it extracts the underlying dataset's configuration,
        handles the `versioned` flag and removes unnecessary metadata. It also
        ensures the embedded dataset's configuration is appropriately flattened
        or transformed.

        If the dataset has a version key, it sets the `versioned` flag in the
        configuration.

        Removes the `metadata` key from the configuration if present.

        Returns:
            A dictionary containing the dataset's type and initialization arguments.
        """
        return_config: dict[str, Any] = {
            f"{TYPE_KEY}": f"{type(self).__module__}.{type(self).__name__}"
        }

        if self._init_args:  # type: ignore[attr-defined]
            return_config.update(self._init_args)  # type: ignore[attr-defined]

        if type(self).__name__ == "CachedDataset":
            cached_ds = return_config.pop("dataset")
            cached_ds_return_config: dict[str, Any] = {}
            if isinstance(cached_ds, dict):
                cached_ds_return_config = cached_ds
            elif isinstance(cached_ds, AbstractDataset):
                cached_ds_return_config = cached_ds._init_config()
            if VERSIONED_FLAG_KEY in cached_ds_return_config:
                return_config[VERSIONED_FLAG_KEY] = cached_ds_return_config.pop(
                    VERSIONED_FLAG_KEY
                )
            # Pop metadata from configuration
            cached_ds_return_config.pop("metadata", None)
            return_config["dataset"] = cached_ds_return_config

        # Set `versioned` key if version present in the dataset
        if return_config.pop(VERSION_KEY, None):
            return_config[VERSIONED_FLAG_KEY] = True

        # Pop metadata from configuration
        return_config.pop("metadata", None)

        return return_config

    @property
    def _logger(self) -> logging.Logger:
        return logging.getLogger(__name__)

    @classmethod
    def _load_wrapper(cls, load_func: Callable[[Self], _DO]) -> Callable[[Self], _DO]:
        """Decorate `load_func` with logging and error handling code."""

        @wraps(load_func)
        def load(self: Self) -> _DO:
            self._logger.debug("Loading %s", str(self))

            try:
                return load_func(self)
            except DatasetError:
                raise
            except Exception as exc:
                # This exception handling is by design as the composed datasets
                # can throw any type of exception.
                message = f"Failed while loading data from dataset {self!s}.\n{exc!s}"
                raise DatasetError(message) from exc

        load.__annotations__["return"] = load_func.__annotations__.get("return")
        load.__loadwrapped__ = True  # type: ignore[attr-defined]
        return load

    @classmethod
    def _save_wrapper(
        cls, save_func: Callable[[Self, _DI], None]
    ) -> Callable[[Self, _DI], None]:
        """Decorate `save_func` with logging and error handling code."""

        @wraps(save_func)
        def save(self: Self, data: _DI) -> None:
            if data is None:
                raise DatasetError("Saving 'None' to a 'Dataset' is not allowed")

            try:
                self._logger.debug("Saving %s", str(self))
                save_func(self, data)
            except (DatasetError, FileNotFoundError, NotADirectoryError):
                raise
            except Exception as exc:
                message = f"Failed while saving data to dataset {self!s}.\n{exc!s}"
                raise DatasetError(message) from exc

        save.__annotations__["data"] = save_func.__annotations__.get("data", Any)
        save.__annotations__["return"] = save_func.__annotations__.get("return")
        save.__savewrapped__ = True  # type: ignore[attr-defined]
        return save

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Customizes the behavior of subclasses of AbstractDataset during
        their creation. This method is automatically invoked when a subclass
        of AbstractDataset is defined.

        Decorates the `load` and `save` methods provided by the class.
        If `_load` or `_save` are defined, alias them as a prerequisite.
        """

        # Save the original __init__ method of the subclass
        init_func: Callable = cls.__init__

        @wraps(init_func)
        def new_init(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
            """Executes the original __init__, then save the arguments used
            to initialize the instance.
            """
            # Call the original __init__ method
            init_func(self, *args, **kwargs)
            # Capture and save the arguments passed to the original __init__
            self._init_args = getcallargs(init_func, self, *args, **kwargs)
            self._init_args.pop("self", None)  # removed to prevent recursion

        # Replace the subclass's __init__ with the new_init
        # A hook for subclasses to capture initialization arguments and save them
        # in the AbstractDataset._init_args field
        cls.__init__ = new_init  # type: ignore[method-assign]

        super().__init_subclass__(**kwargs)

        if hasattr(cls, "_load") and not cls._load.__qualname__.startswith("Abstract"):
            cls.load = cls._load  # type: ignore[method-assign]

        if hasattr(cls, "_save") and not cls._save.__qualname__.startswith("Abstract"):
            cls.save = cls._save  # type: ignore[method-assign]

        if hasattr(cls, "load") and not cls.load.__qualname__.startswith("Abstract"):
            cls.load = cls._load_wrapper(  # type: ignore[assignment]
                cls.load
                if not getattr(cls.load, "__loadwrapped__", False)
                else cls.load.__wrapped__  # type: ignore[attr-defined]
            )

        if hasattr(cls, "save") and not cls.save.__qualname__.startswith("Abstract"):
            cls.save = cls._save_wrapper(  # type: ignore[assignment]
                cls.save
                if not getattr(cls.save, "__savewrapped__", False)
                else cls.save.__wrapped__  # type: ignore[attr-defined]
            )

    def _pretty_repr(self, object_description: dict[str, Any]) -> str:
        str_keys = []
        for arg_name, arg_descr in object_description.items():
            if arg_descr is not None:
                descr = pprint.pformat(
                    arg_descr,
                    sort_dicts=False,
                    compact=True,
                    depth=2,
                    width=sys.maxsize,
                )
                str_keys.append(f"{arg_name}={descr}")

        return f"{type(self).__module__}.{type(self).__name__}({', '.join(str_keys)})"

    def __repr__(self) -> str:
        object_description = self._describe()
        if isinstance(object_description, dict) and all(
            isinstance(key, str) for key in object_description
        ):
            return self._pretty_repr(object_description)

        self._logger.warning(
            f"'{type(self).__module__}.{type(self).__name__}' is a subclass of AbstractDataset and it must "
            f"implement the '_describe' method following the signature of AbstractDataset's '_describe'."
        )
        return f"{type(self).__module__}.{type(self).__name__}()"

    @abc.abstractmethod
    def load(self) -> _DO:
        """Loads data by delegation to the provided load method.

        Returns:
            Data returned by the provided load method.

        Raises:
            DatasetError: When underlying load method raises error.

        """
        raise NotImplementedError(
            f"'{self.__class__.__name__}' is a subclass of AbstractDataset and "
            f"it must implement the 'load' method"
        )

    @abc.abstractmethod
    def save(self, data: _DI) -> None:
        """Saves data by delegation to the provided save method.

        Args:
            data: the value to be saved by provided save method.

        Raises:
            DatasetError: when underlying save method raises error.
            FileNotFoundError: when save method got file instead of dir, on Windows.
            NotADirectoryError: when save method got file instead of dir, on Unix.

        """
        raise NotImplementedError(
            f"'{self.__class__.__name__}' is a subclass of AbstractDataset and "
            f"it must implement the 'save' method"
        )

    @abc.abstractmethod
    def _describe(self) -> dict[str, Any]:
        raise NotImplementedError(
            f"'{self.__class__.__name__}' is a subclass of AbstractDataset and "
            f"it must implement the '_describe' method"
        )

    def exists(self) -> bool:
        """Checks whether a dataset's output already exists by calling
        the provided _exists() method.

        Returns:
            Flag indicating whether the output already exists.

        Raises:
            DatasetError: when underlying exists method raises error.

        """
        try:
            self._logger.debug("Checking whether target of %s exists", str(self))
            return self._exists()
        except Exception as exc:
            message = f"Failed during exists check for dataset {self!s}.\n{exc!s}"
            raise DatasetError(message) from exc

    def _exists(self) -> bool:
        self._logger.warning(
            "'exists()' not implemented for '%s'. Assuming output does not exist.",
            self.__class__.__name__,
        )
        return False

    def release(self) -> None:
        """Release any cached data.

        Raises:
            DatasetError: when underlying release method raises error.

        """
        try:
            self._logger.debug("Releasing %s", str(self))
            self._release()
        except Exception as exc:
            message = f"Failed during release for dataset {self!s}.\n{exc!s}"
            raise DatasetError(message) from exc

    def _release(self) -> None:
        pass

    def _copy(self, **overwrite_params: Any) -> AbstractDataset:
        dataset_copy = copy.deepcopy(self)
        for name, value in overwrite_params.items():
            setattr(dataset_copy, name, value)
        return dataset_copy


def generate_timestamp() -> str:
    """Generate the timestamp to be used by versioning.

    Returns:
        String representation of the current timestamp.

    """
    current_ts = datetime.now(tz=timezone.utc).strftime(VERSION_FORMAT)
    return current_ts[:-4] + current_ts[-1:]  # Don't keep microseconds


class Version(namedtuple("Version", ["load", "save"])):
    """This namedtuple is used to provide load and save versions for versioned
    datasets. If ``Version.load`` is None, then the latest available version
    is loaded. If ``Version.save`` is None, then save version is formatted as
    YYYY-MM-DDThh.mm.ss.sssZ of the current timestamp.
    """

    __slots__ = ()


_CONSISTENCY_WARNING = (
    "Save version '{}' did not match load version '{}' for {}. This is strongly "
    "discouraged due to inconsistencies it may cause between 'save' and "
    "'load' operations. Please refrain from setting exact load version for "
    "intermediate datasets where possible to avoid this warning."
)

_DEFAULT_PACKAGES = ["kedro.io.", "kedro_datasets.", ""]


def parse_dataset_definition(
    config: dict[str, Any],
    load_version: str | None = None,
    save_version: str | None = None,
) -> tuple[type[AbstractDataset], dict[str, Any]]:
    """Parse and instantiate a dataset class using the configuration provided.

    Args:
        config: Data set config dictionary. It *must* contain the `type` key
            with fully qualified class name or the class object.
        load_version: Version string to be used for ``load`` operation if
            the dataset is versioned. Has no effect on the dataset
            if versioning was not enabled.
        save_version: Version string to be used for ``save`` operation if
            the dataset is versioned. Has no effect on the dataset
            if versioning was not enabled.

    Raises:
        DatasetError: If the function fails to parse the configuration provided.

    Returns:
        2-tuple: (Dataset class object, configuration dictionary)
    """
    save_version = save_version or generate_timestamp()
    config = copy.deepcopy(config)
    dataset_type = config.pop(TYPE_KEY)

    # This check prevents the use of dataset types with uppercase 'S' in 'Dataset',
    # which is no longer supported as of kedro-datasets 2.0
    if isinstance(dataset_type, str):
        if dataset_type.endswith("Set"):
            warnings.warn(
                f"Since kedro-datasets 2.0, 'Dataset' is spelled with a lowercase 's'. Got '{dataset_type}'.",
                UserWarning,
            )

    class_obj = None
    error_msg = None
    if isinstance(dataset_type, str):
        if len(dataset_type.strip(".")) != len(dataset_type):
            raise DatasetError(
                "'type' class path does not support relative "
                "paths or paths ending with a dot."
            )
        class_paths = (prefix + dataset_type for prefix in _DEFAULT_PACKAGES)

        for class_path in class_paths:
            tmp, error_msg = _load_obj(
                class_path
            )  # Load dataset class, capture the warning

            if tmp is not None:
                class_obj = tmp
                break

        if class_obj is None:  # If no valid class was found, raise an error
            hint = (
                "\nHint: If you are trying to use a dataset from `kedro-datasets`, "
                "make sure that the package is installed in your current environment. "
                "You can do so by running `pip install kedro-datasets` or "
                "`pip install kedro-datasets[<dataset-group>]` to install `kedro-datasets` along with "
                "related dependencies for the specific dataset group."
            )
            default_error_msg = f"Class '{dataset_type}' not found, is this a typo?"
            raise DatasetError(f"{error_msg if error_msg else default_error_msg}{hint}")

    if not class_obj:
        class_obj = dataset_type

    if not issubclass(class_obj, AbstractDataset):
        raise DatasetError(
            f"Dataset type '{class_obj.__module__}.{class_obj.__qualname__}' "
            f"is invalid: all dataset types must extend 'AbstractDataset'."
        )

    if VERSION_KEY in config:
        # remove "version" key so that it's not passed
        # to the "unversioned" dataset constructor
        message = (
            "'%s' attribute removed from dataset configuration since it is a "
            "reserved word and cannot be directly specified"
        )
        logging.getLogger(__name__).warning(message, VERSION_KEY)
        del config[VERSION_KEY]

    # dataset is either versioned explicitly by the user or versioned is set to true by default
    # on the dataset
    if config.pop(VERSIONED_FLAG_KEY, False) or getattr(
        class_obj, VERSIONED_FLAG_KEY, False
    ):
        config[VERSION_KEY] = Version(load_version, save_version)

    return class_obj, config


def _load_obj(class_path: str) -> tuple[Any | None, str | None]:
    """Try to load an object from a fully-qualified class path.

    Raises:
        DatasetError: If the class is listed in `__all__` but cannot be loaded,
        indicating missing dependencies.

    Returns:
        A tuple of (class object or None, error message or None).
    """
    mod_path, _, class_name = class_path.rpartition(".")
    # Check if the module exists
    try:
        available_classes = load_obj(f"{mod_path}.__all__")
        # ModuleNotFoundError: When `load_obj` can't find `mod_path` (e.g `kedro.io.pandas`)
        #                      this is because we try a combination of all prefixes.
        # AttributeError: When `load_obj` manages to load `mod_path` but it doesn't have an
        #                 `__all__` attribute -- either because it's a custom or a kedro.io dataset
    except (ModuleNotFoundError, AttributeError, ValueError):
        available_classes = None

    try:
        class_obj = load_obj(class_path)
    except ValueError as exc:
        return (
            None,
            f"{exc}. Invalid dataset path: '{class_path}'. Please check if it's correct.",
        )
    except AttributeError as exc:
        # Raise if class exists in __all__ but failed to load (likely due to missing deps)
        if available_classes and class_name in available_classes:
            raise DatasetError(
                f"{exc}. Please see the documentation on how to "
                f"install relevant dependencies for {class_path}:\n"
                f"https://docs.kedro.org/en/stable/kedro_project_setup/"
                f"dependencies.html#install-dependencies-related-to-the-data-catalog"
            ) from exc

        return (
            None,
            f"Dataset '{class_name}' not found in '{mod_path}'. "
            f"Make sure the dataset name is correct.",
        )
    except ModuleNotFoundError as exc:
        return (
            None,
            f"{exc}. Please install the missing dependencies for {class_path}:\n"
            f"https://docs.kedro.org/en/stable/kedro_project_setup/"
            f"dependencies.html#install-dependencies-related-to-the-data-catalog",
        )

    return class_obj, None


def _local_exists(filepath: str) -> bool:  # SKIP_IF_NO_SPARK
    return Path(filepath).exists()


class AbstractVersionedDataset(AbstractDataset[_DI, _DO], abc.ABC):
    """
    ``AbstractVersionedDataset`` is the base class for all versioned dataset
    implementations.

    All datasets that implement versioning should extend this
    abstract class and implement the methods marked as abstract.

    Example:
    ``` python
    from pathlib import Path, PurePosixPath
    import pandas as pd
    from kedro.io import AbstractVersionedDataset


    class MyOwnDataset(AbstractVersionedDataset):
        def __init__(self, filepath, version, param1, param2=True):
            super().__init__(PurePosixPath(filepath), version)
            self._param1 = param1
            self._param2 = param2

        def load(self) -> pd.DataFrame:
            load_path = self._get_load_path()
            return pd.read_csv(load_path)

        def save(self, df: pd.DataFrame) -> None:
            save_path = self._get_save_path()
            df.to_csv(str(save_path))

        def _exists(self) -> bool:
            path = self._get_load_path()
            return Path(path.as_posix()).exists()

        def _describe(self):
            return dict(version=self._version, param1=self._param1, param2=self._param2)
    ```

    Example catalog.yml specification:
    ``` yaml
    my_dataset:
        type: <path-to-my-own-dataset>.MyOwnDataset
        filepath: data/01_raw/my_data.csv
        versioned: true
        param1: <param1-value> # param1 is a required argument
        # param2 will be True by default
    ```
    """

    def __init__(
        self,
        filepath: PurePosixPath,
        version: Version | None,
        exists_function: Callable[[str], bool] | None = None,
        glob_function: Callable[[str], list[str]] | None = None,
    ):
        """Creates a new instance of ``AbstractVersionedDataset``.

        Args:
            filepath: Filepath in POSIX format to a file.
            version: If specified, should be an instance of
                ``kedro.io.core.Version``. If its ``load`` attribute is
                None, the latest version will be loaded. If its ``save``
                attribute is None, save version will be autogenerated.
            exists_function: Function that is used for determining whether
                a path exists in a filesystem.
            glob_function: Function that is used for finding all paths
                in a filesystem, which match a given pattern.
        """
        self._filepath = filepath
        self._version = version
        self._exists_function = exists_function or _local_exists
        self._glob_function = glob_function or iglob
        # 1 entry for load version, 1 for save version
        self._version_cache = Cache(maxsize=2)  # type: Cache

    # 'key' is set to prevent cache key overlapping for load and save:
    # https://cachetools.readthedocs.io/en/stable/#cachetools.cachedmethod
    @cachedmethod(cache=attrgetter("_version_cache"), key=partial(hashkey, "load"))
    def _fetch_latest_load_version(self) -> str:
        # When load version is unpinned, fetch the most recent existing
        # version from the given path.
        pattern = str(self._get_versioned_path("*"))
        try:
            version_paths = sorted(self._glob_function(pattern), reverse=True)
        except Exception as exc:
            message = (
                f"Did not find any versions for {self}. This could be "
                f"due to insufficient permission. Exception: {exc}"
            )
            raise VersionNotFoundError(message) from exc
        most_recent = next(
            (path for path in version_paths if self._exists_function(path)), None
        )
        if not most_recent:
            message = f"Did not find any versions for {self}"
            raise VersionNotFoundError(message)
        return PurePath(most_recent).parent.name

    # 'key' is set to prevent cache key overlapping for load and save:
    # https://cachetools.readthedocs.io/en/stable/#cachetools.cachedmethod
    @cachedmethod(cache=attrgetter("_version_cache"), key=partial(hashkey, "save"))
    def _fetch_latest_save_version(self) -> str:
        """Generate and cache the current save version"""
        return generate_timestamp()

    def resolve_load_version(self) -> str | None:
        """Compute the version the dataset should be loaded with."""
        if not self._version:
            return None
        if self._version.load:
            return self._version.load  # type: ignore[no-any-return]
        return self._fetch_latest_load_version()

    def _get_load_path(self) -> PurePosixPath:
        if not self._version:
            # When versioning is disabled, load from original filepath
            return self._filepath

        load_version = self.resolve_load_version()
        return self._get_versioned_path(load_version)  # type: ignore[arg-type]

    def resolve_save_version(self) -> str | None:
        """Compute the version the dataset should be saved with."""
        if not self._version:
            return None
        if self._version.save:
            return self._version.save  # type: ignore[no-any-return]
        return self._fetch_latest_save_version()

    def _get_save_path(self) -> PurePosixPath:
        if not self._version:
            # When versioning is disabled, return original filepath
            return self._filepath

        save_version = self.resolve_save_version()
        versioned_path = self._get_versioned_path(save_version)  # type: ignore[arg-type]

        if self._exists_function(str(versioned_path)):
            raise DatasetError(
                f"Save path '{versioned_path}' for {self!s} must not exist if "
                f"versioning is enabled."
            )

        return versioned_path

    def _get_versioned_path(self, version: str) -> PurePosixPath:
        return self._filepath / version / self._filepath.name

    @classmethod
    def _save_wrapper(
        cls, save_func: Callable[[Self, _DI], None]
    ) -> Callable[[Self, _DI], None]:
        """Decorate `save_func` with logging and error handling code."""

        @wraps(save_func)
        def save(self: Self, data: _DI) -> None:
            self._version_cache.clear()
            save_version = (
                self.resolve_save_version()
            )  # Make sure last save version is set
            try:
                super()._save_wrapper(save_func)(self, data)
            except (FileNotFoundError, NotADirectoryError) as err:
                # FileNotFoundError raised in Win, NotADirectoryError raised in Unix
                _default_version = "YYYY-MM-DDThh.mm.ss.sssZ"
                raise DatasetError(
                    f"Cannot save versioned dataset '{self._filepath.name}' to "
                    f"'{self._filepath.parent.as_posix()}' because a file with the same "
                    f"name already exists in the directory. This is likely because "
                    f"versioning was enabled on a dataset already saved previously. Either "
                    f"remove '{self._filepath.name}' from the directory or manually "
                    f"convert it into a versioned dataset by placing it in a versioned "
                    f"directory (e.g. with default versioning format "
                    f"'{self._filepath.as_posix()}/{_default_version}/{self._filepath.name}"
                    f"')."
                ) from err

            load_version = self.resolve_load_version()
            if load_version != save_version:
                warnings.warn(
                    _CONSISTENCY_WARNING.format(save_version, load_version, str(self))
                )
                self._version_cache.clear()

        return save

    def exists(self) -> bool:
        """Checks whether a dataset's output already exists by calling
        the provided _exists() method.

        Returns:
            Flag indicating whether the output already exists.

        Raises:
            DatasetError: when underlying exists method raises error.

        """
        self._logger.debug("Checking whether target of %s exists", str(self))
        try:
            return self._exists()
        except VersionNotFoundError:
            return False
        except Exception as exc:  # SKIP_IF_NO_SPARK
            message = f"Failed during exists check for dataset {self!s}.\n{exc!s}"
            raise DatasetError(message) from exc

    def _release(self) -> None:
        super()._release()
        self._version_cache.clear()


def get_protocol_and_path(
    filepath: str | os.PathLike, version: Version | None = None
) -> tuple[str, str]:
    """Parses filepath on protocol and path.

    .. warning::
        Versioning is not supported for HTTP protocols.

    Args:
        filepath: raw filepath e.g.: ``gcs://bucket/test.json``.
        version: instance of ``kedro.io.core.Version`` or None.

    Returns:
        Protocol and path.

    Raises:
        DatasetError: when protocol is http(s) and version is not None.
    """
    options_dict = _parse_filepath(str(filepath))
    path = options_dict["path"]
    protocol = options_dict["protocol"]

    if protocol in HTTP_PROTOCOLS:
        if version is not None:
            raise DatasetError(
                "Versioning is not supported for HTTP protocols. "
                "Please remove the `versioned` flag from the dataset configuration."
            )
        path = path.split(PROTOCOL_DELIMITER, 1)[-1]

    return protocol, path


def get_filepath_str(raw_path: PurePath, protocol: str) -> str:
    """Returns filepath. Returns full filepath (with protocol) if protocol is HTTP(s).

    Args:
        raw_path: filepath without protocol.
        protocol: protocol.

    Returns:
        Filepath string.
    """
    path = raw_path.as_posix()
    if protocol in HTTP_PROTOCOLS:
        path = "".join((protocol, PROTOCOL_DELIMITER, path))
    return path


def validate_on_forbidden_chars(**kwargs: Any) -> None:
    """Validate that string values do not include white-spaces or ;"""
    for key, value in kwargs.items():
        if " " in value or ";" in value:
            raise DatasetError(
                f"Neither white-space nor semicolon are allowed in '{key}'."
            )


def is_parameter(dataset_name: str) -> bool:
    """Check if dataset is a parameter."""
    return dataset_name.startswith("params:") or dataset_name == "parameters"


_C = TypeVar("_C")
_DS = TypeVar("_DS")


@runtime_checkable
class CatalogProtocol(Protocol[_C, _DS]):
    _datasets: dict[str, _DS]

    def __repr__(self) -> str:
        """Returns the canonical string representation of the object."""
        ...

    def __contains__(self, ds_name: str) -> bool:
        """Check if a dataset is in the catalog."""
        ...

    def keys(self) -> List[str]:  # noqa: UP006
        """List all dataset names registered in the catalog."""
        ...

    def values(self) -> List[_DS]:  # noqa: UP006
        """List all datasets registered in the catalog."""
        ...

    def items(self) -> List[tuple[str, _DS]]:  # noqa: UP006
        """List all dataset names and datasets registered in the catalog."""
        ...

    def __iter__(self) -> Iterator[str]:
        """Returns an iterator for the object."""
        ...

    def __getitem__(self, ds_name: str) -> _DS:
        """Get a dataset by name from an internal collection of datasets."""
        ...

    def __setitem__(self, key: str, value: Any) -> None:
        """Adds dataset using the given key as a dataset name and the provided data as the value."""
        ...

    @classmethod
    def from_config(cls, catalog: dict[str, dict[str, Any]] | None) -> _C:
        """Create a catalog instance from configuration."""
        ...

    def get(self, key: str, fallback_to_runtime_pattern: bool = False) -> _DS | None:
        """Get a dataset by name from an internal collection of datasets."""
        ...

    def save(self, name: str, data: Any) -> None:
        """Save data to a registered dataset."""
        ...

    def load(self, name: str, version: str | None = None) -> Any:
        """Load data from a registered dataset."""
        ...

    def release(self, name: str) -> None:
        """Release any cached data associated with a dataset."""
        ...

    def confirm(self, name: str) -> None:
        """Confirm a dataset by its name."""
        ...

    def exists(self, name: str) -> bool:
        """Checks whether registered dataset exists by calling its `exists()` method."""
        ...


@runtime_checkable
class SharedMemoryCatalogProtocol(CatalogProtocol, Protocol):
    def set_manager_datasets(self, manager: SyncManager) -> None: ...

    def validate_catalog(self) -> None: ...



"""This module implements Kedro session responsible for project lifecycle."""

from __future__ import annotations

import getpass
import logging
import logging.config
import os
import subprocess
import sys
import traceback
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any

import click

from kedro import __version__ as kedro_version
from kedro.framework.hooks import _create_hook_manager
from kedro.framework.hooks.manager import _register_hooks, _register_hooks_entry_points
from kedro.framework.project import (
    pipelines,
    settings,
    validate_settings,
)
from kedro.io.core import generate_timestamp
from kedro.io.data_catalog import SharedMemoryDataCatalog
from kedro.runner import AbstractRunner, ParallelRunner, SequentialRunner
from kedro.utils import find_kedro_project

if TYPE_CHECKING:
    from collections.abc import Iterable

    from kedro.config import AbstractConfigLoader
    from kedro.framework.context import KedroContext
    from kedro.framework.session.store import BaseSessionStore


def _describe_git(project_path: Path) -> dict[str, dict[str, Any]]:
    path = str(project_path)
    try:
        res = subprocess.check_output(  # noqa: S603
            ["git", "rev-parse", "--short", "HEAD"],  # noqa: S607
            cwd=path,
            stderr=subprocess.STDOUT,
        )
        git_data: dict[str, Any] = {"commit_sha": res.decode().strip()}
        git_status_res = subprocess.check_output(  # noqa: S603
            ["git", "status", "--short"],  # noqa: S607
            cwd=path,
            stderr=subprocess.STDOUT,
        )
        git_data["dirty"] = bool(git_status_res.decode().strip())

    # `subprocess.check_output()` raises `NotADirectoryError` on Windows
    except Exception:
        logger = logging.getLogger(__name__)
        logger.debug("Unable to git describe %s", path)
        logger.debug(traceback.format_exc())
        return {}

    return {"git": git_data}


def _jsonify_cli_context(ctx: click.core.Context) -> dict[str, Any]:
    return {
        "args": ctx.args,
        "params": ctx.params,
        "command_name": ctx.command.name,
        "command_path": " ".join(["kedro"] + sys.argv[1:]),
    }


class KedroSessionError(Exception):
    """``KedroSessionError`` raised by ``KedroSession``
    in the case that multiple runs are attempted in one session.
    """

    pass


class KedroSession:
    """``KedroSession`` is the object that is responsible for managing the lifecycle
    of a Kedro run. Use `KedroSession.create()` as
    a context manager to construct a new KedroSession with session data
    provided (see the example below).



    Example:
    ``` python
    from kedro.framework.session import KedroSession
    from kedro.framework.startup import bootstrap_project
    from pathlib import Path

    # If you are creating a session outside of a Kedro project (i.e. not using
    # `kedro run` or `kedro jupyter`), you need to run `bootstrap_project` to
    # let Kedro find your configuration.
    bootstrap_project(Path("<project_root>"))
    with KedroSession.create() as session:
        session.run()
    ```
    """

    def __init__(
        self,
        session_id: str,
        package_name: str | None = None,
        project_path: Path | str | None = None,
        save_on_close: bool = False,
        conf_source: str | None = None,
    ):
        self._project_path = Path(
            project_path or find_kedro_project(Path.cwd()) or Path.cwd()
        ).resolve()
        self.session_id = session_id
        self.save_on_close = save_on_close
        self._package_name = package_name
        self._store = self._init_store()
        self._run_called = False

        hook_manager = _create_hook_manager()
        _register_hooks(hook_manager, settings.HOOKS)
        _register_hooks_entry_points(hook_manager, settings.DISABLE_HOOKS_FOR_PLUGINS)
        self._hook_manager = hook_manager

        self._conf_source = conf_source or str(
            self._project_path / settings.CONF_SOURCE
        )

    @classmethod
    def create(
        cls,
        project_path: Path | str | None = None,
        save_on_close: bool = True,
        env: str | None = None,
        runtime_params: dict[str, Any] | None = None,
        conf_source: str | None = None,
    ) -> KedroSession:
        """Create a new instance of ``KedroSession`` with the session data.

        Args:
            project_path: Path to the project root directory. Default is
                current working directory Path.cwd().
            save_on_close: Whether or not to save the session when it's closed.
            conf_source: Path to a directory containing configuration
            env: Environment for the KedroContext.
            runtime_params: Optional dictionary containing extra project parameters
                for underlying KedroContext. If specified, will update (and therefore
                take precedence over) the parameters retrieved from the project
                configuration.

        Returns:
            A new ``KedroSession`` instance.
        """
        validate_settings()

        session = cls(
            project_path=project_path,
            session_id=generate_timestamp(),
            save_on_close=save_on_close,
            conf_source=conf_source,
        )

        # have to explicitly type session_data otherwise mypy will complain
        # possibly related to this: https://github.com/python/mypy/issues/1430
        session_data: dict[str, Any] = {
            "project_path": session._project_path,
            "session_id": session.session_id,
        }

        ctx = click.get_current_context(silent=True)
        if ctx:
            session_data["cli"] = _jsonify_cli_context(ctx)

        env = env or os.getenv("KEDRO_ENV")
        if env:
            session_data["env"] = env

        if runtime_params:
            session_data["runtime_params"] = runtime_params

        try:
            session_data["username"] = getpass.getuser()
        except Exception as exc:
            logging.getLogger(__name__).debug(
                "Unable to get username. Full exception: %s", exc
            )

        session_data.update(**_describe_git(session._project_path))
        session._store.update(session_data)

        return session

    def _init_store(self) -> BaseSessionStore:
        store_class = settings.SESSION_STORE_CLASS
        classpath = f"{store_class.__module__}.{store_class.__qualname__}"
        store_args = deepcopy(settings.SESSION_STORE_ARGS)
        store_args.setdefault("path", (self._project_path / "sessions").as_posix())
        store_args["session_id"] = self.session_id

        try:
            return store_class(**store_args)  # type: ignore[no-any-return]
        except TypeError as err:
            raise ValueError(
                f"\n{err}.\nStore config must only contain arguments valid "
                f"for the constructor of '{classpath}'."
            ) from err
        except Exception as err:
            raise ValueError(
                f"\n{err}.\nFailed to instantiate session store of type '{classpath}'."
            ) from err

    def _log_exception(self, exc_type: Any, exc_value: Any, exc_tb: Any) -> None:
        type_ = [] if exc_type.__module__ == "builtins" else [exc_type.__module__]
        type_.append(exc_type.__qualname__)

        exc_data = {
            "type": ".".join(type_),
            "value": str(exc_value),
            "traceback": traceback.format_tb(exc_tb),
        }
        self._store["exception"] = exc_data

    @property
    def _logger(self) -> logging.Logger:
        return logging.getLogger(__name__)

    @property
    def store(self) -> dict[str, Any]:
        """Return a copy of internal store."""
        return dict(self._store)

    def load_context(self) -> KedroContext:
        """An instance of the project context."""
        env = self.store.get("env")
        runtime_params = self.store.get("runtime_params")
        config_loader = self._get_config_loader()
        context_class = settings.CONTEXT_CLASS
        context = context_class(
            package_name=self._package_name,
            project_path=self._project_path,
            config_loader=config_loader,
            env=env,
            runtime_params=runtime_params,
            hook_manager=self._hook_manager,
        )
        self._hook_manager.hook.after_context_created(context=context)

        return context  # type: ignore[no-any-return]

    def _get_config_loader(self) -> AbstractConfigLoader:
        """An instance of the config loader."""
        env = self.store.get("env")
        runtime_params = self.store.get("runtime_params")

        config_loader_class = settings.CONFIG_LOADER_CLASS
        return config_loader_class(  # type: ignore[no-any-return]
            conf_source=self._conf_source,
            env=env,
            runtime_params=runtime_params,
            **settings.CONFIG_LOADER_ARGS,
        )

    def close(self) -> None:
        """Close the current session and save its store to disk
        if `save_on_close` attribute is True.
        """
        if self.save_on_close:
            self._store.save()

    def __enter__(self) -> KedroSession:
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, tb_: Any) -> None:
        if exc_type:
            self._log_exception(exc_type, exc_value, tb_)
        self.close()

    def run(  # noqa: PLR0913
        self,
        pipeline_name: str | None = None,
        tags: Iterable[str] | None = None,
        runner: AbstractRunner | None = None,
        node_names: Iterable[str] | None = None,
        from_nodes: Iterable[str] | None = None,
        to_nodes: Iterable[str] | None = None,
        from_inputs: Iterable[str] | None = None,
        to_outputs: Iterable[str] | None = None,
        load_versions: dict[str, str] | None = None,
        namespaces: Iterable[str] | None = None,
        only_missing_outputs: bool = False,
    ) -> dict[str, Any]:
        """Runs the pipeline with a specified runner.

        Args:
            pipeline_name: Name of the pipeline that is being run.
            tags: An optional list of node tags which should be used to
                filter the nodes of the ``Pipeline``. If specified, only the nodes
                containing *any* of these tags will be run.
            runner: An optional parameter specifying the runner that you want to run
                the pipeline with.
            node_names: An optional list of node names which should be used to
                filter the nodes of the ``Pipeline``. If specified, only the nodes
                with these names will be run.
            from_nodes: An optional list of node names which should be used as a
                starting point of the new ``Pipeline``.
            to_nodes: An optional list of node names which should be used as an
                end point of the new ``Pipeline``.
            from_inputs: An optional list of input datasets which should be
                used as a starting point of the new ``Pipeline``.
            to_outputs: An optional list of output datasets which should be
                used as an end point of the new ``Pipeline``.
            load_versions: An optional flag to specify a particular dataset
                version timestamp to load.
            namespaces: The namespaces of the nodes that are being run.
            only_missing_outputs: Run only nodes with missing outputs.
        Raises:
            ValueError: If the named or `__default__` pipeline is not
                defined by `register_pipelines`.
            Exception: Any uncaught exception during the run will be re-raised
                after being passed to ``on_pipeline_error`` hook.
            KedroSessionError: If more than one run is attempted to be executed during
                a single session.
        Returns:
            Dictionary with pipeline outputs, where keys are dataset names
            and values are dataset objects.
        """
        # Report project name
        self._logger.info("Kedro project %s", self._project_path.name)

        if self._run_called:
            raise KedroSessionError(
                "A run has already been completed as part of the"
                " active KedroSession. KedroSession has a 1-1 mapping with"
                " runs, and thus only one run should be executed per session."
            )

        session_id = self.store["session_id"]
        save_version = session_id
        runtime_params = self.store.get("runtime_params") or {}
        context = self.load_context()

        name = pipeline_name or "__default__"

        try:
            pipeline = pipelines[name]
        except KeyError as exc:
            raise ValueError(
                f"Failed to find the pipeline named '{name}'. "
                f"It needs to be generated and returned "
                f"by the 'register_pipelines' function."
            ) from exc

        filtered_pipeline = pipeline.filter(
            tags=tags,
            from_nodes=from_nodes,
            to_nodes=to_nodes,
            node_names=node_names,
            from_inputs=from_inputs,
            to_outputs=to_outputs,
            node_namespaces=namespaces,
        )

        record_data = {
            "session_id": session_id,
            "project_path": self._project_path.as_posix(),
            "env": context.env,
            "kedro_version": kedro_version,
            "tags": tags,
            "from_nodes": from_nodes,
            "to_nodes": to_nodes,
            "node_names": node_names,
            "from_inputs": from_inputs,
            "to_outputs": to_outputs,
            "load_versions": load_versions,
            "runtime_params": runtime_params,
            "pipeline_name": pipeline_name,
            "namespaces": namespaces,
            "runner": getattr(runner, "__name__", str(runner)),
            "only_missing_outputs": only_missing_outputs,
        }

        runner = runner or SequentialRunner()
        if not isinstance(runner, AbstractRunner):
            raise KedroSessionError(
                "KedroSession expect an instance of Runner instead of a class."
                "Have you forgotten the `()` at the end of the statement?"
            )

        catalog_class = (
            SharedMemoryDataCatalog
            if isinstance(runner, ParallelRunner)
            else settings.DATA_CATALOG_CLASS
        )

        catalog = context._get_catalog(
            catalog_class=catalog_class,
            save_version=save_version,
            load_versions=load_versions,
        )

        # Run the runner
        hook_manager = self._hook_manager
        hook_manager.hook.before_pipeline_run(
            run_params=record_data, pipeline=filtered_pipeline, catalog=catalog
        )
        try:
            run_result = runner.run(
                filtered_pipeline,
                catalog,
                hook_manager,
                run_id=session_id,
                only_missing_outputs=only_missing_outputs,
            )
            self._run_called = True
        except Exception as error:
            hook_manager.hook.on_pipeline_error(
                error=error,
                run_params=record_data,
                pipeline=filtered_pipeline,
                catalog=catalog,
            )
            raise

        hook_manager.hook.after_pipeline_run(
            run_params=record_data,
            run_result=run_result,
            pipeline=filtered_pipeline,
            catalog=catalog,
        )
        return run_result



"""This module provides a set of helper functions being used across different components
of kedro package.
"""

import importlib
import logging
import os
import re
from pathlib import Path
from typing import Any
from urllib.parse import urlsplit

_PYPROJECT = "pyproject.toml"

# Protocols
HTTP_PROTOCOLS = ("http", "https")
CLOUD_PROTOCOLS = (
    "abfs",
    "abfss",
    "adl",
    "gcs",
    "gdrive",
    "gs",
    "oci",
    "oss",
    "s3",
    "s3a",
    "s3n",
)
PROTOCOL_DELIMITER = "://"


def _parse_filepath(filepath: str) -> dict[str, str]:
    """Split filepath on protocol and path. Based on `fsspec.utils.infer_storage_options`.

    Args:
        filepath: Either local absolute file path or URL (s3://bucket/file.csv)

    Returns:
        Parsed filepath.
    """
    if (
        re.match(r"^[a-zA-Z]:[\\/]", filepath)
        or re.match(r"^[a-zA-Z0-9]+://", filepath) is None
    ):
        return {"protocol": "file", "path": filepath}

    parsed_path = urlsplit(filepath)
    protocol = parsed_path.scheme or "file"

    if protocol in HTTP_PROTOCOLS:
        return {"protocol": protocol, "path": filepath}

    path = parsed_path.path
    if protocol == "file":
        windows_path = re.match(r"^/([a-zA-Z])[:|]([\\/].*)$", path)
        if windows_path:
            path = ":".join(windows_path.groups())

    if parsed_path.query:
        path = f"{path}?{parsed_path.query}"
    if parsed_path.fragment:
        path = f"{path}#{parsed_path.fragment}"

    options = {"protocol": protocol, "path": path}

    if parsed_path.netloc and protocol in CLOUD_PROTOCOLS:
        host_with_port = parsed_path.netloc.rsplit("@", 1)[-1]
        host = host_with_port.rsplit(":", 1)[0]
        options["path"] = host + options["path"]
        # - Azure Data Lake Storage Gen2 URIs can store the container name in the
        #   'username' field of a URL (@ syntax), so we need to add it to the path
        # - Oracle Cloud Infrastructure (OCI) Object Storage filesystem (ocifs) also
        #   uses the @ syntax for I/O operations: "oci://bucket@namespace/path_to_file"
        if protocol in ["abfss", "oci"] and parsed_path.username:
            options["path"] = parsed_path.username + "@" + options["path"]

    return options


def load_obj(obj_path: str, default_obj_path: str = "") -> Any:
    """Extract an object from a given path.

    Args:
        obj_path: Path to an object to be extracted, including the object name.
        default_obj_path: Default object path.

    Returns:
        Extracted object.

    Raises:
        AttributeError: When the object does not have the given named attribute.

    """
    obj_path_list = obj_path.rsplit(".", 1)
    obj_path = obj_path_list.pop(0) if len(obj_path_list) > 1 else default_obj_path
    obj_name = obj_path_list[0]
    module_obj = importlib.import_module(obj_path)
    return getattr(module_obj, obj_name)


def _is_databricks() -> bool:
    """Evaluate if the current run environment is Databricks or not.

    Useful to tailor environment-dependent activities like Kedro magic commands
    or logging features that Databricks doesn't support.

    Returns:
        True if run environment is Databricks, otherwise False.
    """
    return "DATABRICKS_RUNTIME_VERSION" in os.environ


def is_kedro_project(project_path: str | Path) -> bool:
    """Evaluate if a given path is a root directory of a Kedro project or not.

    Args:
        project_path: Path to be tested for being a root of a Kedro project.

    Returns:
        True if a given path is a root directory of a Kedro project, otherwise False.
    """
    metadata_file = Path(project_path).expanduser().resolve() / _PYPROJECT
    if not metadata_file.is_file():
        return False

    try:
        return "[tool.kedro]" in metadata_file.read_text(encoding="utf-8")
    except Exception:
        return False


def find_kedro_project(current_dir: Path) -> Any:  # pragma: no cover
    """Given a path, find a Kedro project associated with it.

    Can be:
        - Itself, if a path is a root directory of a Kedro project.
        - One of its parents, if self is not a Kedro project but one of the parent path is.
        - None, if neither self nor any parent path is a Kedro project.

    Returns:
        Kedro project associated with a given path,
        or None if no relevant Kedro project is found.
    """
    paths_to_check = [current_dir, *list(current_dir.parents)]
    for parent_dir in paths_to_check:
        if is_kedro_project(parent_dir):
            return parent_dir
    return None


def _has_rich_handler(logger: logging.Logger | None = None) -> bool:
    """Returns true if the logger has a RichHandler attached."""
    if not logger:
        logger = logging.getLogger()  # User root by default
    try:
        from rich.logging import RichHandler
    except ImportError:
        return False
    return any(isinstance(handler, RichHandler) for handler in logger.handlers)


def _format_rich(value: str, markup: str) -> str:
    """Format string with rich markup"""
    return f"[{markup}]{value}[/{markup}]"



"""``DataCatalog`` stores instances of ``AbstractDataset`` implementations to
provide ``load`` and ``save`` capabilities from anywhere in the program. To
use a ``DataCatalog``, you need to instantiate it with a dictionary of datasets.
Then it will act as a single point of reference for your calls, relaying load and
save functions to the underlying datasets.
"""

from __future__ import annotations

import logging
import re
from multiprocessing.reduction import ForkingPickler
from pickle import PicklingError
from typing import TYPE_CHECKING, Any, ClassVar, Iterator, List  # noqa: UP035

from kedro.io.cached_dataset import CachedDataset
from kedro.io.catalog_config_resolver import CatalogConfigResolver
from kedro.io.core import (
    TYPE_KEY,
    AbstractDataset,
    AbstractVersionedDataset,
    CatalogProtocol,
    DatasetError,
    DatasetNotFoundError,
    Version,
    VersionAlreadyExistsError,
    generate_timestamp,
    parse_dataset_definition,
)
from kedro.io.memory_dataset import MemoryDataset, _is_memory_dataset
from kedro.io.shared_memory_dataset import SharedMemoryDataset
from kedro.utils import _format_rich, _has_rich_handler

if TYPE_CHECKING:
    from multiprocessing.managers import SyncManager


class _LazyDataset:
    """
    A helper class to store `AbstractDataset` configuration and materialize the dataset object.

    This class acts as a placeholder for a dataset, storing its configuration and versioning
    information. The dataset is only materialized (i.e., instantiated) when explicitly requested,
    which helps improve performance by deferring the creation of datasets until they are needed.

    Attributes:
        name (str): The name of the dataset.
        config (dict[str, Any]): The configuration dictionary for the dataset.
        load_version (str | None): The version of the dataset to load, if applicable.
        save_version (str | None): The version of the dataset to save, if applicable.

    Example:
    ```python

        from kedro.io.data_catalog import _LazyDataset
        dataset_config = {"type": "pandas.CSVDataset", "filepath": "example.csv"}
        lazy_dataset = _LazyDataset(
            name="example_dataset",
            config=dataset_config,
            load_version="2023-01-01T00.00.00",
            save_version="2023-01-02T00.00.00",
        )
        print(lazy_dataset)
        # kedro_datasets.pandas.csv_dataset.CSVDataset
        materialized_dataset = lazy_dataset.materialize()
        print(materialized_dataset)
        # CSVDataset(filepath=example.csv, load_args={}, protocol=file, save_args={'index': False})
    ```
    """

    def __init__(
        self,
        name: str,
        config: dict[str, Any],
        load_version: str | None = None,
        save_version: str | None = None,
    ):
        """
        Initialize a `_LazyDataset` instance.

        Args:
            name: The name of the dataset.
            config: The configuration dictionary for the dataset.
            load_version: The version of the dataset to load, if applicable.
            save_version: The version of the dataset to save, if applicable.

        Example:
        ```python

            from kedro.io.data_catalog import _LazyDataset
            dataset_config = {"type": "pandas.CSVDataset", "filepath": "example.csv"}
            lazy_dataset = _LazyDataset(
                name="example_dataset",
                config=dataset_config,
                load_version="2023-01-01T00.00.00",
                save_version="2023-01-02T00.00.00",
            )
            print(lazy_dataset)
            # kedro_datasets.pandas.csv_dataset.CSVDataset
        ```
        """
        self.name = name
        self.config = config
        self.load_version = load_version
        self.save_version = save_version

    def __repr__(self) -> str:
        """
        Return a string representation of the `_LazyDataset`.

        The representation includes the fully qualified class name of the dataset type
        as defined in the configuration.

        Returns:
            The fully qualified class name of the dataset type.

        Example:
        ````python

            from kedro.io.data_catalog import _LazyDataset
            dataset_config = {"type": "pandas.CSVDataset", "filepath": "example.csv"}
            lazy_dataset = _LazyDataset(name="example_dataset", config=dataset_config)
            print(repr(lazy_dataset))
            # kedro_datasets.pandas.csv_dataset.CSVDataset
        ```
        """
        class_type, _ = parse_dataset_definition(self.config)
        return f"{class_type.__module__}.{class_type.__qualname__}"

    def materialize(self) -> AbstractDataset:
        """
        Materialize the `_LazyDataset` into an `AbstractDataset` instance.

        This method uses the stored configuration and versioning information to create
        an instance of the dataset. The dataset is instantiated using the `from_config`
        factory method of `AbstractDataset`.

        Returns:
            The instantiated dataset object.

        Example:
        ``` python

            from kedro.io.data_catalog import _LazyDataset
            dataset_config = {"type": "pandas.CSVDataset", "filepath": "example.csv"}
            lazy_dataset = _LazyDataset(
                name="example_dataset",
                config=dataset_config,
                load_version="2023-01-01T00.00.00",
                save_version="2023-01-02T00.00.00",
            )
            materialized_dataset = lazy_dataset.materialize()
            print(materialized_dataset)
            # CSVDataset(filepath=example.csv, load_args={}, protocol=file, save_args={'index': False})
        ```
        """
        return AbstractDataset.from_config(
            self.name, self.config, self.load_version, self.save_version
        )


class DataCatalog(CatalogProtocol):
    """
    A centralized registry for managing datasets in a Kedro project.

    The `DataCatalog` provides a unified interface for loading and saving datasets,
    enabling seamless interaction with various data sources and formats. It supports
    features such as lazy loading, versioning, and dynamic dataset creation using dataset
    factory patterns.

    This class is the core component of Kedro's data management system, allowing
    datasets to be defined, accessed, and manipulated in a consistent and reusable way.

    Attributes:
        default_runtime_patterns (ClassVar): A dictionary defining the default runtime pattern
            for datasets of type `kedro.io.MemoryDataset`.
        _datasets: A dictionary of fully initialized datasets.
        _lazy_datasets: A dictionary of `_LazyDataset` instances for deferred initialization.
        _load_versions: A mapping of dataset names to specific versions to load.
        _save_version: The global version string for saving datasets.
        _config_resolver: Resolves dataset factory patterns and configurations.

    Example:
    ``` python
        from kedro.io import DataCatalog, MemoryDataset

        # Define datasets
        datasets = {
            "cars": MemoryDataset(data={"type": "car", "capacity": 5}),
            "planes": MemoryDataset(data={"type": "jet", "capacity": 200}),
        }

        # Initialize the catalog
        catalog = DataCatalog(datasets=datasets)

        # Load data
        cars_data = catalog.load("cars")
        print(cars_data)

        # Save data
        catalog.save("planes", {"type": "propeller", "capacity": 100})
        planes_data = catalog.load("planes")
        print(planes_data)
    ```
    """

    default_runtime_patterns: ClassVar = {
        "{default}": {"type": "kedro.io.MemoryDataset"}
    }

    def __init__(
        self,
        datasets: dict[str, AbstractDataset] | None = None,
        config_resolver: CatalogConfigResolver | None = None,
        load_versions: dict[str, str] | None = None,
        save_version: str | None = None,
    ) -> None:
        """Initializes a ``DataCatalog`` to manage datasets with loading, saving, and versioning capabilities.

        This catalog combines datasets passed directly via the `datasets` argument and dynamic datasets
        resolved from config (e.g., from YAML).

        If a dataset name is present in both `datasets` and the resolved config, the dataset from `datasets`
        takes precedence. A warning is logged, and the config-defined dataset is skipped and removed from
        the internal config.

        Args:
            datasets: A dictionary of dataset names and dataset instances.
            config_resolver: An instance of CatalogConfigResolver to resolve dataset factory patterns and configurations.
            load_versions: A mapping between dataset names and versions
                to load. Has no effect on datasets without enabled versioning.
            save_version: Version string to be used for ``save`` operations
                by all datasets with enabled versioning. It must: a) be a
                case-insensitive string that conforms with operating system
                filename limitations, b) always return the latest version when
                sorted in lexicographical order.

        Example:
        ``` python

            from kedro.io import DataCatalog, MemoryDataset
            from kedro_datasets.pandas import CSVDataset

            # Define datasets
            datasets = {
                "cars": CSVDataset(filepath="cars.csv"),
                "planes": MemoryDataset(data={"type": "jet", "capacity": 200}),
            }

            # Initialize the catalog
            catalog = DataCatalog(
                datasets=datasets,
                load_versions={"cars": "2023-01-01T00.00.00"},
                save_version="2023-01-02T00.00.00",
            )

            print(catalog)
        ```
        """
        self._config_resolver = config_resolver or CatalogConfigResolver(
            default_runtime_patterns=self.default_runtime_patterns
        )
        self._datasets: dict[str, AbstractDataset] = datasets or {}
        self._lazy_datasets: dict[str, _LazyDataset] = {}
        self._load_versions, self._save_version = self._validate_versions(
            datasets, load_versions or {}, save_version
        )

        self._use_rich_markup = _has_rich_handler()

        for ds_name in list(self._config_resolver.config):
            if ds_name in self._datasets:
                self._logger.warning(
                    f"Cannot register dataset '{ds_name}' from config: a dataset with the same name "
                    f"was already provided in the `datasets` argument."
                )
                self._config_resolver.config.pop(ds_name)
            else:
                self._add_from_config(ds_name, self._config_resolver.config[ds_name])

    @property
    def config_resolver(self) -> CatalogConfigResolver:
        """
        Get the `CatalogConfigResolver` instance associated with this `DataCatalog`.

        The `CatalogConfigResolver` is responsible for resolving dataset factory patterns
        and configurations dynamically.

        Returns:
            CatalogConfigResolver: The configuration resolver for the catalog.

        Example:
        ``` python

            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            resolver = catalog.config_resolver
            print(resolver)
        ```
        """
        return self._config_resolver

    def __repr__(self) -> str:
        """
        Return a string representation of the `DataCatalog`.

        The representation includes both lazy and fully initialized datasets
        in the catalog.

        Returns:
            A string representation of the catalog.

        Example:
        ```python

            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            print(repr(catalog))
            # "example: kedro.io.memory_dataset.MemoryDataset()"
        ```
        """
        combined = self._lazy_datasets | self._datasets
        lines = []
        for key, dataset in combined.items():
            lines.append(f"'{key}': {dataset!r}")
        return "\n".join(lines)

    def __contains__(self, dataset_name: str) -> bool:
        """
        Check if a dataset is registered in the catalog
        or matches a dataset/user catch all pattern

        Args:
            dataset_name: The name of the dataset to check.

        Returns:
            True if the dataset is registered or matches a pattern, False otherwise.

        Example:
        ```python

            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            "example" in catalog
            # True
            "nonexistent" in catalog
            # False
        ```
        """
        return (
            dataset_name in self._datasets
            or dataset_name in self._lazy_datasets
            or self._config_resolver.match_dataset_pattern(dataset_name) is not None
            or self._config_resolver.match_user_catch_all_pattern(dataset_name)
            is not None
        )

    def __eq__(
        self, other: Any
    ) -> bool:  # Any type to silence warning on mkdocs strict mode
        """
        Compare two catalogs based on materialized datasets, lazy datasets and all dataset factory patterns.

        Args:
            other: Another `DataCatalog` instance to compare.

        Returns:
            True if the catalogs are equivalent, False otherwise.

        Example:
        ```python
            catalog1 = DataCatalog(datasets={"example": MemoryDataset()})
            catalog2 = DataCatalog(datasets={"example": MemoryDataset()})
            catalog1 == catalog2
            # False
        ```
        """
        return (
            self._datasets,
            self._lazy_datasets,
            self._config_resolver.list_patterns(),
        ) == (
            other._datasets,
            other._lazy_datasets,
            other.config_resolver.list_patterns(),
        )

    def keys(self) -> List[str]:  # noqa: UP006
        """
        List all dataset names registered in the catalog, including both materialized and lazy datasets.

        Returns:
            A list of all dataset names.

        Example:
        ```python
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            catalog.keys()
            # ['example']
        ```
        """
        return list(self._lazy_datasets.keys()) + list(self._datasets.keys())

    def values(self) -> List[AbstractDataset]:  # noqa: UP006
        """
        Retrieve all datasets registered in the catalog.

        This method returns a list of all dataset instances currently registered
        in the catalog, including both materialized and lazy datasets.

        Returns:
            A list of dataset instances.

        Example:
        ```python
            from kedro.io import DataCatalog, MemoryDataset
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            dataset_values = catalog.values()
            print(dataset_values)
            # [kedro.io.memory_dataset.MemoryDataset()]
        ```
        """
        return [self[key] for key in self]

    def items(self) -> List[tuple[str, AbstractDataset]]:  # noqa: UP006
        """
        Retrieve all dataset names and their corresponding dataset instances.

        This method returns a list of tuples, where each tuple contains a dataset
        name and its corresponding dataset instance.

        Returns:
            A list of tuples containing dataset names and their corresponding dataset instances.

        Example:
        ```python

            from kedro.io import DataCatalog, MemoryDataset
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            dataset_items = catalog.items()
            print(dataset_items)
            # [('example', kedro.io.memory_dataset.MemoryDataset())]
        ```
        """
        return [(key, self[key]) for key in self]

    def __iter__(self) -> Iterator[str]:
        """
        Iterate over all dataset names in the catalog, including both materialized and lazy datasets.

        Yields:
            Dataset names.

        Example:
        ```python
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            for dataset_name in catalog:
                print(dataset_name)
            # example
        ```
        """
        yield from self.keys()

    def __getitem__(self, ds_name: str) -> AbstractDataset:
        """
        Get a dataset by name from the catalog.

        If a dataset is not materialized but matches dataset_pattern or user_catch_all_pattern
        by default or runtime_patterns if fallback_to_runtime_pattern is enabled,
        it is instantiated and added to the catalog first, then returned.

        Args:
            ds_name: The name of the dataset.

        Returns:
            The dataset instance.

        Raises:
            DatasetNotFoundError: If the dataset is not found.

        Example:
        ```python
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            dataset = catalog["example"]
            print(dataset)
            # MemoryDataset()
        ```
        """
        dataset = self.get(ds_name)
        if dataset is None:
            raise DatasetNotFoundError(f"Dataset '{ds_name}' not found in the catalog")
        return dataset

    def __setitem__(self, key: str, value: Any) -> None:
        """Registers a dataset or raw data into the catalog using the specified name.

        If the name already exists, the dataset will be replaced and relevant internal mappings
        (lazy datasets, config, versions) will be cleared to avoid conflicts.

        The value can either be raw data or a Kedro dataset (i.e., an instance of a class
        inheriting from ``AbstractDataset``). If raw data is provided, it will be automatically
        wrapped in a ``MemoryDataset`` before being added to the catalog.

        Args:
            key: Name of the dataset.
            value: Either an instance of `AbstractDataset`, `_LazyDataset`, or raw data (e.g., DataFrame, list).
                If raw data is provided, it is automatically wrapped as a `MemoryDataset`.

        Example:
        ```python

            from kedro_datasets.pandas import CSVDataset
            import pandas as pd
            df = pd.DataFrame({"col1": [1, 2], "col2": [4, 5], "col3": [5, 6]})
            catalog = DataCatalog()
            catalog["data_df"] = df  # Add raw data as a MemoryDataset

            assert catalog.load("data_df").equals(df)

            csv_dataset = CSVDataset(filepath="test.csv")
            csv_dataset.save(df)
            catalog["data_csv_dataset"] = csv_dataset  # Add a dataset instance

            assert catalog.load("data_csv_dataset").equals(df)
        ```
        """
        if key in self._datasets or key in self._lazy_datasets:
            self._logger.warning("Replacing dataset '%s'", key)
            self._datasets.pop(key, None)
            self._lazy_datasets.pop(key, None)
            self._config_resolver.config.pop(key, None)
            self._load_versions.pop(key, None)
        if isinstance(value, AbstractDataset):
            self._load_versions, self._save_version = self._validate_versions(
                {key: value}, self._load_versions, self._save_version
            )
            self._datasets[key] = value
        elif isinstance(value, _LazyDataset):
            self._lazy_datasets[key] = value
        else:
            self._logger.debug(f"Adding input data {key} as a default MemoryDataset")
            self._datasets[key] = MemoryDataset(data=value)  # type: ignore[abstract]

    def __len__(self) -> int:
        """
        Get the number of datasets registered in the catalog, including both materialized and lazy datasets.

        Returns:
            The number of datasets.

        Example:
        ```python
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            len(catalog)
            # 1
        ```
        """
        return len(self.keys())

    def get(
        self,
        key: str,
        fallback_to_runtime_pattern: bool = False,
        version: Version | None = None,
    ) -> AbstractDataset | None:
        """Get a dataset by name from an internal collection of datasets.

        If a dataset is not materialized but matches dataset_pattern or user_catch_all_pattern
        by default or runtime_patterns if fallback_to_runtime_pattern is enabled
        it is instantiated and added to the catalog first, then returned.

        Args:
            key: A dataset name.
            fallback_to_runtime_pattern: Whether to use runtime_pattern to resolve dataset.
            version: Optional argument to get a specific version of the dataset.

        Returns:
            The dataset instance if found, otherwise None.

        Example:
        ```python
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            dataset = catalog.get("example")
            print(dataset)
            # MemoryDataset()

            missing = catalog.get("nonexistent")
            print(missing)
            # None
        ```
        """
        if key not in self and not fallback_to_runtime_pattern:
            return None

        if not (key in self._datasets or key in self._lazy_datasets):
            ds_config = self._config_resolver.resolve_pattern(key)
            if ds_config:
                self._add_from_config(key, ds_config)

        lazy_dataset = self._lazy_datasets.pop(key, None)
        if lazy_dataset:
            self[key] = lazy_dataset.materialize()

        dataset = self._datasets[key]

        if version and isinstance(dataset, AbstractVersionedDataset):
            # we only want to return a similar-looking dataset,
            # not modify the one stored in the current catalog
            dataset = dataset._copy(_version=version)

        return dataset

    def _ipython_key_completions_(self) -> list[str]:
        return self.keys()

    @property
    def _logger(self) -> logging.Logger:
        return logging.getLogger(__name__)

    @classmethod
    def from_config(
        cls,
        catalog: dict[str, dict[str, Any]] | None,
        credentials: dict[str, dict[str, Any]] | None = None,
        load_versions: dict[str, str] | None = None,
        save_version: str | None = None,
    ) -> DataCatalog:
        """Create a ``DataCatalog`` instance from configuration. This is a
        factory method used to provide developers with a way to instantiate
        ``DataCatalog`` with configuration parsed from configuration files.

        Args:
            catalog: A dictionary whose keys are the dataset names and
                the values are dictionaries with the constructor arguments
                for classes implementing ``AbstractDataset``. The dataset
                class to be loaded is specified with the key ``type`` and their
                fully qualified class name. All ``kedro.io`` dataset can be
                specified by their class name only, i.e. their module name
                can be omitted.
            credentials: A dictionary containing credentials for different
                datasets. Use the ``credentials`` key in a ``AbstractDataset``
                to refer to the appropriate credentials as shown in the example
                below.
            load_versions: A mapping between dataset names and versions
                to load. Has no effect on datasets without enabled versioning.
            save_version: Version string to be used for ``save`` operations
                by all datasets with enabled versioning. It must: a) be a
                case-insensitive string that conforms with operating system
                filename limitations, b) always return the latest version when
                sorted in lexicographical order.

        Returns:
            An instantiated ``DataCatalog`` containing all specified
            datasets, created and ready to use.

        Raises:
            DatasetNotFoundError: When `load_versions` refers to a dataset that doesn't
                exist in the catalog.

        Example:
        ```python
            config = {
                "cars": {
                    "type": "pandas.CSVDataset",
                    "filepath": "cars.csv",
                    "save_args": {"index": False}
                },
                "boats": {
                    "type": "pandas.CSVDataset",
                    "filepath": "s3://aws-bucket-name/boats.csv",
                    "credentials": "boats_credentials",
                    "save_args": {"index": False}
                }
            }
            credentials = {
                "boats_credentials": {
                    "client_kwargs": {
                        "aws_access_key_id": "<your key id>",
                        "aws_secret_access_key": "<your secret>"
                    }
                }
            }

            catalog = DataCatalog.from_config(config, credentials)
            df = catalog.load("cars")

            catalog.save("boats", df)
        ```
        """
        catalog = catalog or {}
        config_resolver = CatalogConfigResolver(
            catalog, credentials, cls.default_runtime_patterns
        )
        save_version = save_version or generate_timestamp()
        load_versions = load_versions or {}

        missing_keys = [
            ds_name
            for ds_name in load_versions
            if not (
                ds_name in config_resolver.config
                or config_resolver.match_dataset_pattern(ds_name)
            )
        ]
        if missing_keys:
            raise DatasetNotFoundError(
                f"'load_versions' keys [{', '.join(sorted(missing_keys))}] "
                f"are not found in the catalog."
            )

        return cls(
            load_versions=load_versions,
            save_version=save_version,
            config_resolver=config_resolver,
        )

    def to_config(
        self,
    ) -> tuple[
        dict[str, dict[str, Any]],
        dict[str, dict[str, Any]],
        dict[str, str | None],
        str | None,
    ]:
        """Converts the `DataCatalog` instance into a configuration format suitable for
        serialization. This includes datasets, credentials, and versioning information.

        This method is only applicable to catalogs that contain datasets initialized with static, primitive
        parameters. For example, it will work fine if one passes credentials as dictionary to
        `GBQQueryDataset` but not as `google.auth.credentials.Credentials` object. See
        https://github.com/kedro-org/kedro-plugins/issues/950 for the details.

        Returns:
            A tuple containing:
                catalog: A dictionary mapping dataset names to their unresolved configurations,
                    excluding in-memory datasets.

                credentials: A dictionary of unresolved credentials extracted from dataset configurations.

                load_versions: A dictionary mapping dataset names to specific versions to be loaded,
                    or `None` if no version is set.

                save_version: A global version identifier for saving datasets, or `None` if not specified.

        Example:
        ```python
            from kedro.io import DataCatalog
            from kedro_datasets.pandas import CSVDataset

            cars = CSVDataset(
                filepath="cars.csv",
                load_args=None,
                save_args={"index": False}
            )
            catalog = DataCatalog(datasets={'cars': cars})

            config, credentials, load_versions, save_version = catalog.to_config()

            new_catalog = DataCatalog.from_config(config, credentials, load_versions, save_version)
        ```
        """
        catalog: dict[str, dict[str, Any]] = {}
        credentials: dict[str, dict[str, Any]] = {}
        load_versions: dict[str, str | None] = {}

        for ds_name, ds in self._lazy_datasets.items():
            if _is_memory_dataset(ds.config.get(TYPE_KEY, "")):
                continue
            unresolved_config, unresolved_credentials = (
                self._config_resolver._unresolve_credentials(ds_name, ds.config)
            )
            catalog[ds_name] = unresolved_config
            credentials.update(unresolved_credentials)
            load_versions[ds_name] = self._load_versions.get(ds_name, None)

        for ds_name, ds in self._datasets.items():  # type: ignore[assignment]
            if _is_memory_dataset(ds):  # type: ignore[arg-type]
                continue
            resolved_config = ds._init_config()  # type: ignore[attr-defined]
            unresolved_config, unresolved_credentials = (
                self._config_resolver._unresolve_credentials(ds_name, resolved_config)
            )
            catalog[ds_name] = unresolved_config
            credentials.update(unresolved_credentials)
            load_versions[ds_name] = self._load_versions.get(ds_name, None)

        return catalog, credentials, load_versions, self._save_version

    @staticmethod
    def _validate_dataset_config(ds_name: str, ds_config: Any) -> None:
        """
        Validate the configuration of a dataset in the catalog.

        This method ensures that the dataset configuration is a dictionary and contains
        the required "type" key. If the configuration is invalid, it raises a `DatasetError`
        with a helpful error message.

        Args:
            ds_name: The name of the dataset being validated.
            ds_config: The configuration of the dataset to validate.

        Raises:
            DatasetError: If the dataset configuration is not a dictionary or if the
                "type" key is missing from the configuration.

        Example:
        ```python
            config = {
                "example_dataset": {
                    "type": "pandas.CSVDataset",
                    "filepath": "example.csv",
                }
            }
            DataCatalog._validate_dataset_config("example_dataset", config["example_dataset"])
            # No error raised

            invalid_config = {"example_dataset": {"filepath": "example.csv"}}
            DataCatalog._validate_dataset_config("example_dataset", invalid_config["example_dataset"])
            # Raises DatasetError: 'type' is missing from dataset catalog configuration.
        ```
        """
        if not isinstance(ds_config, dict):
            raise DatasetError(
                f"Catalog entry '{ds_name}' is not a valid dataset configuration. "
                "\nHint: If this catalog entry is intended for variable interpolation, "
                "make sure that the key is preceded by an underscore."
            )

        if "type" not in ds_config:
            raise DatasetError(
                f"An exception occurred when parsing config for dataset '{ds_name}':\n"
                "'type' is missing from dataset catalog configuration."
                "\nHint: If this catalog entry is intended for variable interpolation, "
                "make sure that the top level key is preceded by an underscore."
            )

    def _add_from_config(self, ds_name: str, ds_config: dict[str, Any]) -> None:
        """
        Create a `_LazyDataset` instance from the provided configuration and add it to the catalog.

        This method validates the dataset configuration, creates a `_LazyDataset` instance,
        and registers it in the catalog. The `_LazyDataset` is a placeholder for a dataset
        that will be materialized only when accessed.

        Args:
            ds_name: The name of the dataset to be added.
            ds_config: The configuration dictionary for the dataset.

        Raises:
            DatasetError: If the provided dataset configuration is invalid (e.g., missing
                required keys like "type" or improperly formatted).

        Example:
        ```python
            catalog = DataCatalog()
            dataset_config = {"type": "pandas.CSVDataset", "filepath": "example.csv"}
            catalog._add_from_config("example_dataset", dataset_config)
            "example_dataset" in catalog
            # True
        ```
        """
        self._validate_dataset_config(ds_name, ds_config)
        ds = _LazyDataset(
            ds_name,
            ds_config,
            self._load_versions.get(ds_name),
            self._save_version,
        )

        self.__setitem__(ds_name, ds)

    def filter(
        self,
        name_regex: re.Pattern[str] | str | None = None,
        type_regex: re.Pattern[str] | str | None = None,
        by_type: type | list[type] | None = None,
    ) -> List[str]:  # noqa: UP006
        """Filter dataset names registered in the catalog based on name and/or type.

        This method allows filtering datasets by their names and/or types. Regular expressions
        should be precompiled before passing them to `name_regex` or `type_regex`, but plain
        strings are also supported.

        Args:
            name_regex: Optional compiled regex pattern or string to filter dataset names.
            type_regex: Optional compiled regex pattern or string to filter dataset types.
                The provided regex is matched against the full dataset type path, for example:
                `kedro_datasets.pandas.parquet_dataset.ParquetDataset`.
            by_type: Optional dataset type(s) to filter by. This performs an instance type check
                rather than a regex match. It can be a single dataset type or a list of types.

        Returns:
            A list of dataset names that match the filtering criteria.

        Example:
        ```python
            import re
            catalog = DataCatalog()
            # get datasets where the substring 'raw' is present
            raw_data = catalog.filter(name_regex='raw')
            # get datasets where names start with 'model_' (precompiled regex)
            model_datasets = catalog.filter(name_regex=re.compile('^model_'))
            # get datasets of a specific type using type_regex
            csv_datasets = catalog.filter(type_regex='pandas.excel_dataset.ExcelDataset')
            # get datasets where names contain 'train' and type matches 'CSV' in the path
            catalog.filter(name_regex="train", type_regex="CSV")
            # get datasets where names include 'data' and are of a specific type
            from kedro_datasets.pandas import SQLQueryDataset
            catalog.filter(name_regex="data", by_type=SQLQueryDataset)
            # get datasets where names include 'data' and are of multiple specific types
            from kedro.io import MemoryDataset
            catalog.filter(name_regex="data", by_type=[MemoryDataset, SQLQueryDataset])
        ```

        """
        filtered = self.keys()

        # Apply name filter if specified
        if name_regex:
            filtered = [
                ds_name for ds_name in filtered if re.search(name_regex, ds_name)
            ]

        # Apply type filters if specified
        by_type_set = set()
        if by_type:
            if not isinstance(by_type, list):
                by_type = [by_type]
            for _type in by_type:
                by_type_set.add(f"{_type.__module__}.{_type.__qualname__}")

        if by_type_set or type_regex:
            filtered_types = []
            for ds_name in filtered:
                # Retrieve the dataset type
                str_type = self.get_type(ds_name) or ""
                # Match against type_regex and apply by_type filtering
                if (not type_regex or re.search(type_regex, str_type)) and (
                    not by_type_set or str_type in by_type_set
                ):
                    filtered_types.append(ds_name)

            return filtered_types

        return filtered

    def get_type(self, ds_name: str) -> str | None:
        """
        Access dataset type without adding resolved dataset to the catalog.

        Args:
            ds_name: The name of the dataset whose type is to be retrieved.

        Returns:
            The fully qualified type of the dataset (e.g., `kedro.io.memory_dataset.MemoryDataset`)
            or None if dataset does not match dataset_patterns or user_catch_all_pattern.

        Example:
        ```python
            from kedro.io import DataCatalog, MemoryDataset
            catalog = DataCatalog(datasets={"example": MemoryDataset()})
            dataset_type = catalog.get_type("example")
            print(dataset_type)
            # kedro.io.memory_dataset.MemoryDataset

            missing_type = catalog.get_type("nonexistent")
            print(missing_type)
            # None
        ```
        """
        if ds_name not in self:
            return None

        if ds_name not in self._datasets and ds_name not in self._lazy_datasets:
            ds_config = self._config_resolver.resolve_pattern(ds_name)
            return str(_LazyDataset(ds_name, ds_config))

        if ds_name in self._lazy_datasets:
            return str(self._lazy_datasets[ds_name])

        class_type = type(self._datasets[ds_name])
        return f"{class_type.__module__}.{class_type.__qualname__}"

    def save(self, ds_name: str, data: Any) -> None:
        """Save data to a registered dataset.

        Args:
            ds_name: The name of the dataset to be saved.
            data: A data object to be saved as configured in the registered
                dataset.

        Raises:
            DatasetNotFoundError: When a dataset with the given name
                has not yet been registered.

        Example:
        ```python
            import pandas as pd

            from kedro.io import DataCatalog
            from kedro_datasets.pandas import CSVDataset

            cars = CSVDataset(filepath="cars.csv", load_args=None, save_args={"index": False})
            catalog = DataCatalog(datasets={'cars': cars})
            df = pd.DataFrame({'col1': [1, 2], 'col2': [4, 5], 'col3': [5, 6]})
            catalog.save("cars", df)
        ```
        """
        dataset = self[ds_name]

        self._logger.info(
            "Saving data to %s (%s)...",
            _format_rich(ds_name, "dark_orange") if self._use_rich_markup else ds_name,
            type(dataset).__name__,
            extra={"markup": True},
        )

        dataset.save(data)

    def load(self, ds_name: str, version: str | None = None) -> Any:
        """Loads a registered dataset.

        Args:
            ds_name: The name of the dataset to be loaded.
            version: Optional argument for concrete data version to be loaded.
                Works only with versioned datasets.

        Returns:
            The loaded data as configured.

        Raises:
            DatasetNotFoundError: When a dataset with the given name
                has not yet been registered.

        Example:
        ```python
            from kedro.io import DataCatalog
            from kedro_datasets.pandas import CSVDataset
            cars = CSVDataset(filepath="cars.csv", load_args=None, save_args={"index": False})
            catalog = DataCatalog(datasets={'cars': cars})
            df = catalog.load("cars")
        ```
        """
        load_version = Version(version, None) if version else None
        dataset = self.get(ds_name, version=load_version)

        if dataset is None:
            error_msg = f"Dataset '{ds_name}' not found in the catalog"
            raise DatasetNotFoundError(error_msg)

        self._logger.info(
            "Loading data from %s (%s)...",
            _format_rich(ds_name, "dark_orange") if self._use_rich_markup else ds_name,
            type(dataset).__name__,
            extra={"markup": True},
        )

        return dataset.load()

    def release(self, ds_name: str) -> None:
        """Release any cached data associated with a dataset
        Args:
            ds_name: A dataset to be checked.
        Raises:
            DatasetNotFoundError: When a dataset with the given name
                has not yet been registered.

        Example:
        ```python
            catalog = DataCatalog(datasets={"example": MemoryDataset(data=[1, 2, 3])})
            catalog.release("example")
        ```
        """
        dataset = self[ds_name]

        dataset.release()

    def confirm(self, ds_name: str) -> None:
        """Confirm a dataset by its name.
        Args:
            ds_name: Name of the dataset.
        Raises:
            DatasetNotFoundError: When a dataset with the given name
                has not yet been registered
            DatasetError: When the dataset does not have `confirm` method.
        """
        self._logger.info("Confirming dataset '%s'", ds_name)

        dataset = self[ds_name]

        if hasattr(dataset, "confirm"):
            dataset.confirm()
        else:
            raise DatasetError(f"Dataset '{ds_name}' does not have 'confirm' method")

    def exists(self, ds_name: str) -> bool:
        """Checks whether registered dataset exists by calling its `exists()`
        method. Raises a warning and returns False if `exists()` is not
        implemented.

        Args:
            ds_name: A dataset to be checked.

        Returns:
            Whether the dataset and its output exist.

        Example:
        ```python
            catalog = DataCatalog(datasets={"example": MemoryDataset(data=[1, 2, 3])})
            catalog.exists("example")
            True
        ```
        """
        dataset = self.get(ds_name)
        return dataset.exists() if dataset else False

    @staticmethod
    def _validate_versions(
        datasets: dict[str, AbstractDataset] | None,
        load_versions: dict[str, str],
        save_version: str | None,
    ) -> tuple[dict[str, str], str | None]:
        """Validates and synchronises dataset versions for loading and saving.

        Ensures consistency of dataset versions across a catalog, particularly
        for versioned datasets. It updates load versions and validates that all
        save versions are consistent.

        Args:
            datasets: A dictionary mapping dataset names to their instances.
                if None, no validation occurs.
            load_versions: A mapping between dataset names and versions
                to load.
            save_version: Version string to be used for ``save`` operations
                by all datasets with versioning enabled.

        Returns:
            Updated ``load_versions`` with load versions specified in the ``datasets``
                and resolved ``save_version``.

        Raises:
            VersionAlreadyExistsError: If a dataset's save version conflicts with
                the catalog's save version.
        """
        if not datasets:
            return load_versions, save_version

        cur_load_versions = load_versions.copy()
        cur_save_version = save_version

        for ds_name, ds in datasets.items():
            cur_ds = ds._dataset if isinstance(ds, CachedDataset) else ds

            if isinstance(cur_ds, AbstractVersionedDataset) and cur_ds._version:
                if cur_ds._version.load:
                    cur_load_versions[ds_name] = cur_ds._version.load
                if cur_ds._version.save:
                    cur_save_version = cur_save_version or cur_ds._version.save
                    if cur_save_version != cur_ds._version.save:
                        raise VersionAlreadyExistsError(
                            f"Cannot add a dataset `{ds_name}` with `{cur_ds._version.save}` save version. "
                            f"Save version set for the catalog is `{cur_save_version}`"
                            f"All datasets in the catalog must have the same save version."
                        )

        return cur_load_versions, cur_save_version


class SharedMemoryDataCatalog(DataCatalog):
    """
    A specialized `DataCatalog` for managing datasets in a shared memory context.

    The `SharedMemoryDataCatalog` extends the base `DataCatalog` to support multiprocessing
    by ensuring that datasets are serializable and synchronized across threads or processes.
    It provides additional functionality for managing shared memory datasets, such as setting
    a multiprocessing manager and validating dataset compatibility with multiprocessing.

    Attributes:
        default_runtime_patterns (ClassVar): A dictionary defining the default runtime pattern
            for datasets of type `kedro.io.SharedMemoryDataset`.

    Example:
    ``` python

        from multiprocessing.managers import SyncManager
        from kedro.io import MemoryDataset
        from kedro.io.data_catalog import SharedMemoryDataCatalog

        # Create a shared memory catalog
        catalog = SharedMemoryDataCatalog(
            datasets={"shared_data": MemoryDataset(data=[1, 2, 3])}
        )

        # Set a multiprocessing manager
        manager = SyncManager()
        manager.start()
        catalog.set_manager_datasets(manager)

        # Validate the catalog for multiprocessing compatibility
        catalog.validate_catalog()
    ```
    """

    default_runtime_patterns: ClassVar = {
        "{default}": {"type": "kedro.io.SharedMemoryDataset"}
    }

    def set_manager_datasets(self, manager: SyncManager) -> None:
        """
        Associate a multiprocessing manager with all shared memory datasets in the catalog.

        This method iterates through all datasets in the catalog and sets the provided
        multiprocessing manager for datasets of type `SharedMemoryDataset`. This ensures
        that these datasets are properly synchronized across threads or processes.

        Args:
            manager: A multiprocessing manager to be associated with
                shared memory datasets.

        Example:
        ```python
            from multiprocessing.managers import SyncManager
            from kedro.io.data_catalog import SharedMemoryDataCatalog
            catalog = SharedMemoryDataCatalog(datasets={"shared_data": MemoryDataset(data=[1, 2, 3])})
            manager = SyncManager()
            manager.start()
            catalog.set_manager_datasets(manager)
            print(catalog)
            # {'shared_data': kedro.io.memory_dataset.MemoryDataset(data='<list>')}
        ```
        """
        for _, ds in self._datasets.items():
            if isinstance(ds, SharedMemoryDataset):
                ds.set_manager(manager)

    def validate_catalog(self) -> None:
        """
        Validate the catalog to ensure all datasets are serializable and compatible with multiprocessing.

        This method checks that all datasets in the catalog are serializable and do not
        include non-proxied memory datasets as outputs. Non-serializable datasets or
        datasets that rely on single-process memory cannot be used in a multiprocessing
        context. If any such datasets are found, an exception is raised with details.

        Raises:
            AttributeError: If any datasets are found to be non-serializable or incompatible
                with multiprocessing.

        Example:
        ```python
            from kedro.io.data_catalog import SharedMemoryDataCatalog

            catalog = SharedMemoryDataCatalog(datasets={"shared_data": MemoryDataset(data=[1, 2, 3])})
            try:
                catalog.validate_catalog()
            except AttributeError as e:
                print(f"Validation failed: {e}")
            # No error
        ```
        """
        unserialisable = []
        for name, dataset in self._datasets.items():
            if getattr(dataset, "_SINGLE_PROCESS", False):  # SKIP_IF_NO_SPARK
                unserialisable.append(name)
                continue
            try:
                ForkingPickler.dumps(dataset)
            except (AttributeError, PicklingError):
                unserialisable.append(name)

        if unserialisable:
            raise AttributeError(
                f"The following datasets cannot be used with multiprocessing: "
                f"{sorted(unserialisable)}\nIn order to utilize multiprocessing you "
                f"need to make sure all datasets are serialisable, i.e. datasets "
                f"should not make use of lambda functions, nested functions, closures "
                f"etc.\nIf you are using custom decorators ensure they are correctly "
                f"decorated using functools.wraps()."
            )



"""``kedro.framework.context`` provides functionality for loading Kedro
project context.
"""

from .catalog_mixins import CatalogCommandsMixin
from .context import (
    KedroContext,
    KedroContextError,
    compose_classes,
)

__all__ = [
    "KedroContext",
    "KedroContextError",
    "CatalogCommandsMixin",
    "compose_classes",
]



import pandas as pd
from kedro_datasets.pandas import CSVDataset

from kedro.io import DataCatalog

base_catalog = {
    f"dataset_{i}": {
        "type": "pandas.CSVDataset",
        "filepath": f"data_{i}.csv",
    } for i in range(1, 1001)
}
# Add datasets with the same filepath for loading
base_catalog.update({
    f"dataset_load_{i}": {
        "type": "pandas.CSVDataset",
        "filepath": "data.csv",
    } for i in range(1, 1001)
})
# Add a factory pattern
base_catalog.update({
    "dataset_factory_{placeholder}": {
        "type": "pandas.CSVDataset",
        "filepath": "data_{placeholder}.csv",
    }
})

runtime_patterns = {
    "{placeholder}": {
        "type": "pandas.CSVDataset",
        "filepath": "{placeholder}.csv",
    }
}

class TimeDataCatalog:
    def setup(self):
        self.catalog = DataCatalog.from_config(base_catalog)
        self.dataframe = pd.DataFrame({"column": [1, 2, 3]})
        self.dataframe.to_csv("data.csv", index=False)
        self.datasets = {
            f"dataset_new_{i}": CSVDataset(filepath="data.csv") for i in range(1, 1001)
        }
        self.parameters = {
            f"param_{i}": i for i in range(1, 1001)
        }

    def time_init(self):
        """Benchmark the time to initialize the catalog"""
        DataCatalog.from_config(base_catalog)

    def time_contains(self):
        """Benchmark the time to check if a dataset exists"""
        for i in range(1,1001):
            f"dataset_{i}" in self.catalog

    def time_getitem(self):
        """Benchmark the time to get a dataset"""
        for i in range(1,1001):
            self.catalog[f"dataset_{i}"]


    def time_get(self):
        """Benchmark the time to get a dataset"""
        for i in range(1,1001):
            self.catalog.get(f"dataset_{i}")

    def time_iter(self):
        """Benchmark the time to iterate over the catalog"""
        for dataset in self.catalog:
            pass

    def time_keys(self):
        """Benchmark the time to get the keys of the catalog"""
        self.catalog.keys()

    def time_values(self):
        """Benchmark the time to get the items of the catalog"""
        self.catalog.values()

    def time_items(self):
        """Benchmark the time to get the items of the catalog"""
        self.catalog.items()

    def time_setitem(self):
        """Benchmark the time to set a dataset"""
        for i in range(1,1001):
            self.catalog[f"dataset_new_{i}"] = CSVDataset(filepath="data.csv")

    def time_setitem_raw(self):
        """Benchmark the time to add a memory dataset"""
        for i in range(1,1001):
            self.catalog[f"param_{i}"] = self.parameters[f"param_{i}"]

    def time_save(self):
        """Benchmark the time to save datasets"""
        for i in range(1,1001):
            self.catalog.save(f"dataset_{i}", self.dataframe)

    def time_load(self):
        """Benchmark the time to load datasets"""
        for i in range(1,1001):
            self.catalog.load(f"dataset_load_{i}")

    def time_exists(self):
        """Benchmark the time to check if datasets exist"""
        for i in range(1,1001):
            self.catalog.exists(f"dataset_{i}")

    def time_release(self):
        """Benchmark the time to release datasets"""
        for i in range(1,1001):
            self.catalog.release(f"dataset_{i}")

    def time_resolve_factory(self):
        """Benchmark the time to resolve factory"""
        for i in range(1,1001):
            self.catalog.get(f"dataset_factory_{i}")



from pathlib import Path

from kedro.templates.project.hooks.utils import setup_template_tools


def main():
    current_dir = Path.cwd()
    requirements_file_path = current_dir / "requirements.txt"
    pyproject_file_path = current_dir / "pyproject.toml"
    python_package_name = "{{ cookiecutter.python_package }}"

    # Get the selected tools from cookiecutter
    selected_tools = "{{ cookiecutter.tools }}"
    example_pipeline = "{{ cookiecutter.example_pipeline }}"

    # Handle template directories and requirements according to selected tools
    setup_template_tools(
        selected_tools,
        requirements_file_path,
        pyproject_file_path,
        python_package_name,
        example_pipeline,
    )


if __name__ == "__main__":
    main()



"""A ``Pipeline`` is a collection of ``Node`` objects which can be executed as
a Directed Acyclic Graph, sequentially or in parallel. The ``Pipeline`` class
offers quick access to input dependencies,
produced outputs and execution order.
"""

from __future__ import annotations

import copy
import difflib
import json
from collections import Counter, defaultdict, deque
from functools import cached_property
from graphlib import CycleError, TopologicalSorter
from itertools import chain
from typing import TYPE_CHECKING, Any
from warnings import warn

import kedro
from kedro.pipeline.node import GroupedNodes, Node, _to_list

from .transcoding import _strip_transcoding, _transcode_split

if TYPE_CHECKING:
    from collections.abc import Iterable, Set


def __getattr__(name: str) -> Any:
    if name == "TRANSCODING_SEPARATOR":
        import warnings

        from kedro.pipeline.transcoding import TRANSCODING_SEPARATOR

        warnings.warn(
            f"{name!r} has been moved to 'kedro.pipeline.transcoding', "
            f"and the alias will be removed in Kedro 1.0.0.",
            kedro.KedroDeprecationWarning,
            stacklevel=2,
        )
        return TRANSCODING_SEPARATOR
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class PipelineError(Exception):
    """Raised when a pipeline is not adapted and integrated
    appropriately using the helper.
    """

    pass


def _is_all_parameters(name: str) -> bool:
    return name == "parameters"


def _is_single_parameter(name: str) -> bool:
    return name.startswith("params:")


def _is_parameter(name: str) -> bool:
    return _is_single_parameter(name) or _is_all_parameters(name)


def _validate_inputs_outputs(
    inputs: Set[str], outputs: Set[str], pipe: Pipeline
) -> None:
    """Safeguards to ensure that:
    - parameters are not specified under inputs
    - inputs are only free inputs
    - outputs do not contain free inputs
    """
    inputs = {_strip_transcoding(k) for k in inputs}
    outputs = {_strip_transcoding(k) for k in outputs}

    if any(_is_parameter(i) for i in inputs):
        raise PipelineError(
            "Parameters should be specified in the 'parameters' argument"
        )

    free_inputs = {_strip_transcoding(i) for i in pipe.inputs()}

    if not inputs <= free_inputs:
        raise PipelineError(
            "Inputs must not be outputs from another node in the same pipeline"
        )

    if outputs & free_inputs:
        raise PipelineError(
            "All outputs must be generated by some node within the pipeline"
        )


def _validate_datasets_exist(
    inputs: Set[str],
    outputs: Set[str],
    parameters: Set[str],
    pipe: Pipeline,
) -> None:
    """Validate that inputs, parameters and outputs map correctly onto the provided nodes."""
    inputs = {_strip_transcoding(k) for k in inputs}
    outputs = {_strip_transcoding(k) for k in outputs}

    existing = {_strip_transcoding(ds) for ds in pipe.datasets()}
    non_existent = (inputs | outputs | parameters) - existing
    if non_existent:
        sorted_non_existent = sorted(non_existent)
        possible_matches = []
        for non_existent_input in sorted_non_existent:
            possible_matches += difflib.get_close_matches(non_existent_input, existing)

        error_msg = f"Failed to map datasets and/or parameters onto the nodes provided: {', '.join(sorted_non_existent)}"
        suggestions = (
            f" - did you mean one of these instead: {', '.join(possible_matches)}"
            if possible_matches
            else ""
        )
        raise PipelineError(error_msg + suggestions)


class OutputNotUniqueError(Exception):
    """Raised when two or more nodes that are part of the same pipeline
    produce outputs with the same name.
    """

    pass


class ConfirmNotUniqueError(Exception):
    """Raised when two or more nodes that are part of the same pipeline
    attempt to confirm the same dataset.
    """

    pass


class Pipeline:
    """A ``Pipeline`` defined as a collection of ``Node`` objects. This class
    treats nodes as part of a graph representation and provides inputs,
    outputs and execution order.
    """

    def __init__(  # noqa: PLR0913
        self,
        nodes: Iterable[Node | Pipeline] | Pipeline,
        *,
        inputs: str | set[str] | dict[str, str] | None = None,
        outputs: str | set[str] | dict[str, str] | None = None,
        parameters: str | set[str] | dict[str, str] | None = None,
        tags: str | Iterable[str] | None = None,
        namespace: str | None = None,
        prefix_datasets_with_namespace: bool = True,
    ):
        """Initialise ``Pipeline`` with a list of ``Node`` instances.

        Args:
            nodes: The iterable of nodes the ``Pipeline`` will be made of. If you
                provide pipelines among the list of nodes, those pipelines will
                be expanded and all their nodes will become part of this
                new pipeline.
            inputs: A name or collection of input names to be exposed as connection points
                to other pipelines upstream. This is optional; if not provided, the
                pipeline inputs are automatically inferred from the pipeline structure.
                When str or set[str] is provided, the listed input names will stay
                the same as they are named in the provided pipeline.
                When dict[str, str] is provided, current input names will be
                mapped to new names.
                Must only refer to the pipeline's free inputs.
            outputs: A name or collection of names to be exposed as connection points
                to other pipelines downstream. This is optional; if not provided, the
                pipeline outputs are automatically inferred from the pipeline structure.
                When str or set[str] is provided, the listed output names will stay
                the same as they are named in the provided pipeline.
                When dict[str, str] is provided, current output names will be
                mapped to new names.
                Can refer to both the pipeline's free outputs, as well as
                intermediate results that need to be exposed.
            parameters: A name or collection of parameters to namespace.
                When str or set[str] are provided, the listed parameter names will stay
                the same as they are named in the provided pipeline.
                When dict[str, str] is provided, current parameter names will be
                mapped to new names.
                The parameters can be specified without the `params:` prefix.
            tags: Optional set of tags to be applied to all the pipeline nodes.
            namespace: A prefix to give to all dataset names,
                except those explicitly named with the `inputs`/`outputs`
                arguments, and parameter references (`params:` and `parameters`).
            prefix_datasets_with_namespace: A flag to specify if the inputs, outputs, and parameters of the nodes
                should be prefixed with the namespace. It is set to True by default. It is
                useful to turn off when namespacing is used for grouping nodes for deployment purposes.

        Raises:
            ValueError:
                When an empty list of nodes is provided, or when not all
                nodes have unique names.
            CircularDependencyError:
                When visiting all the nodes is not
                possible due to the existence of a circular dependency.
            OutputNotUniqueError:
                When multiple ``Node`` instances produce the same output.
            ConfirmNotUniqueError:
                When multiple ``Node`` instances attempt to confirm the same
                dataset.
            PipelineError: When inputs, outputs or parameters are incorrectly
                specified, or they do not exist on the original pipeline.
        Example:
        ``` python
        from kedro.pipeline import Pipeline
        from kedro.pipeline import node

        # In the following scenario first_ds and second_ds
        # are datasets provided by io. Pipeline will pass these
        # datasets to first_node function and provides the result
        # to the second_node as input.


        def first_node(first_ds, second_ds):
            return dict(third_ds=first_ds + second_ds)


        def second_node(third_ds):
            return third_ds


        pipeline = Pipeline(
            [
                node(first_node, ["first_ds", "second_ds"], ["third_ds"]),
                node(second_node, dict(third_ds="third_ds"), "fourth_ds"),
            ]
        )

        pipeline.describe()
        ```
        """
        if isinstance(nodes, Pipeline):
            nodes = nodes.nodes

        if any([inputs, outputs, parameters, namespace]):
            nodes = self._map_nodes(
                nodes=nodes,
                inputs=inputs,
                outputs=outputs,
                parameters=parameters,
                tags=tags,
                namespace=namespace,
                prefix_datasets_with_namespace=prefix_datasets_with_namespace,
            )

        if nodes is None:
            raise ValueError(
                "'nodes' argument of 'Pipeline' is None. It must be an "
                "iterable of nodes and/or pipelines instead."
            )
        nodes_list = list(nodes)  # in case it's a generator
        _validate_duplicate_nodes(nodes_list)

        nodes_chain = list(
            chain.from_iterable(
                [[n] if isinstance(n, Node) else n.nodes for n in nodes_list]
            )
        )
        _validate_transcoded_inputs_outputs(nodes_chain)
        _tags = set(_to_list(tags))

        if _tags:
            tagged_nodes = [n.tag(_tags) for n in nodes_chain]
        else:
            tagged_nodes = nodes_chain

        self._nodes_by_name = {node.name: node for node in tagged_nodes}
        _validate_unique_outputs(tagged_nodes)
        _validate_unique_confirms(tagged_nodes)

        # input -> nodes with input
        self._nodes_by_input: dict[str, set[Node]] = defaultdict(set)
        for node in tagged_nodes:
            for input_ in node.inputs:
                self._nodes_by_input[_strip_transcoding(input_)].add(node)

        # output -> node with output
        self._nodes_by_output: dict[str, Node] = {}
        for node in tagged_nodes:
            for output in node.outputs:
                self._nodes_by_output[_strip_transcoding(output)] = node

        self._nodes = tagged_nodes
        self._toposorter = TopologicalSorter(self.node_dependencies)

        # test for circular dependencies without executing the toposort for efficiency
        try:
            self._toposorter.prepare()
        except CycleError as exc:
            loop = list(set(exc.args[1]))
            message = f"Circular dependencies exist among the following {len(loop)} item(s): {loop}"
            raise CircularDependencyError(message) from exc

        self._toposorted_nodes: list[Node] = []
        self._toposorted_groups: list[list[Node]] = []
        if any(n.namespace for n in self._nodes):
            self._validate_namespaces()

    @cached_property
    def _node_children(self) -> dict[Node, set[Node]]:
        node_children = defaultdict(set)
        for child, parents in self.node_dependencies.items():
            for parent in parents:
                node_children[parent].add(child)
        return node_children

    def _shortest_path(self, start: Node, end: Node) -> list[Node]:
        """Find the shortest path between two nodes in the pipeline, if it exists."""
        queue = deque([(start, [start])])
        visited = set([start])

        while queue:
            node, path = queue.popleft()
            if node == end:
                return path
            for child in self._node_children[node]:
                if child not in visited:
                    visited.add(child)
                    queue.append((child, [*path, child]))
        return []  # pragma: no cover

    def _validate_namespaces(self) -> None:
        node_parents: dict[Node, set[Node]] = self.node_dependencies

        # Here we keep all visited namespaces and the nodes that interrupted them for each node
        # the mapping is as follows:
        # {node: {namespace: {interrupting_node}}}
        seen_namespaces: dict[Node, dict[str, set[Node]]] = defaultdict(dict)
        for node in self.nodes:
            visited: dict[str, set[Node]] = defaultdict(set)
            for parent in node_parents[node]:
                last_namespace = parent.namespace or ""
                curr_namespace = node.namespace or ""
                seen = seen_namespaces[parent]
                # If any part of curr_namespace was visited in paths leading to parent
                interrupted = next(
                    (ns for ns in node.namespace_prefixes if ns in seen),
                    None,
                )
                # The second condition checks if the interruption was already reported during visiting the parent node
                if interrupted and not (last_namespace + ".").startswith(
                    interrupted + "."
                ):
                    start = next(iter(seen[interrupted]))
                    path = [n.name for n in self._shortest_path(start, node)]
                    warn(
                        f"Namespace '{interrupted}' is interrupted by nodes {path[:-1]} and thus invalid.",
                        UserWarning,
                    )
                # All visited namespaces for the current node get updated with the parent's visited namespaces
                for ns, nodes in seen.items():
                    visited[ns].update(nodes)
                # If the current namespace is different from the last namespace and isn't a child namespace,
                # mark the last namespace and all unrelated parent namespaces as visited to detect potential future interruptions
                if (
                    last_namespace
                    and curr_namespace != last_namespace
                    and not curr_namespace.startswith(last_namespace + ".")
                ):
                    for prefix in parent.namespace_prefixes:
                        if not curr_namespace.startswith(prefix):
                            visited[prefix].add(node)

            # Now we have created the visited namespaces for the current node
            seen_namespaces[node] = visited

    def __repr__(self) -> str:  # pragma: no cover
        """Pipeline ([node1, ..., node10 ...], name='pipeline_name')"""
        max_nodes_to_display = 10

        nodes_reprs = [repr(node) for node in self.nodes[:max_nodes_to_display]]
        if len(self.nodes) > max_nodes_to_display:
            nodes_reprs.append("...")
        sep = ",\n"
        nodes_reprs_str = f"[\n{sep.join(nodes_reprs)}\n]" if nodes_reprs else "[]"
        constructor_repr = f"({nodes_reprs_str})"
        return f"{self.__class__.__name__}{constructor_repr}"

    def __add__(self, other: Any) -> Pipeline:
        if not isinstance(other, Pipeline):
            return NotImplemented
        return Pipeline(set(self._nodes + other._nodes))

    def __radd__(self, other: Any) -> Pipeline:
        if isinstance(other, int) and other == 0:
            return self
        return self.__add__(other)

    def __sub__(self, other: Any) -> Pipeline:
        if not isinstance(other, Pipeline):
            return NotImplemented
        return Pipeline(set(self._nodes) - set(other._nodes))

    def __and__(self, other: Any) -> Pipeline:
        if not isinstance(other, Pipeline):
            return NotImplemented
        return Pipeline(set(self._nodes) & set(other._nodes))

    def __or__(self, other: Any) -> Pipeline:
        if not isinstance(other, Pipeline):
            return NotImplemented
        return Pipeline(set(self._nodes + other._nodes))

    def all_inputs(self) -> set[str]:
        """All inputs for all nodes in the pipeline.

        Returns:
            All node input names as a Set.

        """
        return set.union(set(), *(node.inputs for node in self._nodes))

    def all_outputs(self) -> set[str]:
        """All outputs of all nodes in the pipeline.

        Returns:
            All node outputs.

        """
        return set.union(set(), *(node.outputs for node in self._nodes))

    def _remove_intermediates(self, datasets: set[str]) -> set[str]:
        intermediate = {_strip_transcoding(i) for i in self.all_inputs()} & {
            _strip_transcoding(o) for o in self.all_outputs()
        }
        return {d for d in datasets if _strip_transcoding(d) not in intermediate}

    def inputs(self) -> set[str]:
        """The names of free inputs that must be provided at runtime so that
        the pipeline is runnable. Does not include intermediate inputs which
        are produced and consumed by the inner pipeline nodes. Resolves
        transcoded names where necessary.

        Returns:
            The set of free input names needed by the pipeline.

        """
        return self._remove_intermediates(self.all_inputs())

    def outputs(self) -> set[str]:
        """The names of outputs produced when the whole pipeline is run.
        Does not include intermediate outputs that are consumed by
        other pipeline nodes. Resolves transcoded names where necessary.

        Returns:
            The set of final pipeline outputs.

        """
        return self._remove_intermediates(self.all_outputs())

    def datasets(self) -> set[str]:
        """The names of all datasets used by the ``Pipeline``,
        including inputs and outputs.

        Returns:
            The set of all pipeline datasets.

        """
        return self.all_outputs() | self.all_inputs()

    def _transcode_compatible_names(self) -> set[str]:
        return {_strip_transcoding(ds) for ds in self.datasets()}

    def describe(self, names_only: bool = True) -> str:
        """Obtain the order of execution and expected free input variables in
        a loggable pre-formatted string. The order of nodes matches the order
        of execution given by the topological sort.

        Args:
            names_only: The flag to describe names_only pipeline with just
                node names.

        Example:
        ``` python
        pipeline = Pipeline([...])

        logger = logging.getLogger(__name__)

        logger.info(pipeline.describe())
        ```

        After invocation the following will be printed as an info level log
        statement:
        ::

            #### Pipeline execution order ####
            Inputs: C, D

            func1([C]) -> [A]
            func2([D]) -> [B]
            func3([A, D]) -> [E]

            Outputs: B, E
            ##################################

        Returns:
            The pipeline description as a formatted string.

        """

        def set_to_string(set_of_strings: set[str]) -> str:
            """Convert set to a string but return 'None' in case of an empty
            set.
            """
            return ", ".join(sorted(set_of_strings)) if set_of_strings else "None"

        nodes_as_string = "\n".join(
            node.name if names_only else str(node) for node in self.nodes
        )

        str_representation = (
            "#### Pipeline execution order ####\n"
            "Inputs: {0}\n\n"
            "{1}\n\n"
            "Outputs: {2}\n"
            "##################################"
        )

        return str_representation.format(
            set_to_string(self.inputs()), nodes_as_string, set_to_string(self.outputs())
        )

    @cached_property
    def node_dependencies(self) -> dict[Node, set[Node]]:
        """All dependencies of nodes where the first Node has a direct dependency on
        the second Node.

        Returns:
            Dictionary where keys are nodes and values are sets made up of
            their parent nodes. Independent nodes have this as empty sets.
        """
        dependencies: dict[Node, set[Node]] = {node: set() for node in self._nodes}
        for parent in self._nodes:
            for output in parent.outputs:
                for child in self._nodes_by_input[_strip_transcoding(output)]:
                    dependencies[child].add(parent)

        return dependencies

    @cached_property
    def nodes(self) -> list[Node]:
        """Return a list of the pipeline nodes in topological order, i.e. if
        node A needs to be run before node B, it will appear earlier in the
        list.

        Returns:
            The list of all pipeline nodes in topological order.

        """
        if not self._toposorted_nodes:
            self._toposorted_nodes = [n for group in self.grouped_nodes for n in group]

        return list(self._toposorted_nodes)

    @cached_property
    def grouped_nodes(self) -> list[list[Node]]:
        """Return a list of the pipeline nodes in topologically ordered groups,
        i.e. if node A needs to be run before node B, it will appear in an
        earlier group.

        Returns:
            The pipeline nodes in topologically ordered groups.

        """

        if not self._toposorted_groups:
            while self._toposorter:
                group = sorted(self._toposorter.get_ready())
                self._toposorted_groups.append(group)
                self._toposorter.done(*group)

        return [list(group) for group in self._toposorted_groups]

    def group_nodes_by(
        self,
        group_by: str | None = "namespace",
    ) -> list[GroupedNodes]:
        """Return a list of grouped nodes based on the specified strategy.

        Args:
            group_by: Strategy for grouping. Supported values:
                - "namespace": Groups nodes by their top-level namespace.
                - None or "none": No grouping, each node is its own group.

        Returns:
            A list of GroupedNodes instances.
        """
        if group_by is None or group_by.lower() == "none":
            return self._group_by_none()
        if group_by.lower() == "namespace":
            return self._group_by_namespace()
        raise ValueError(f"Unsupported group_by strategy: {group_by}")

    def _group_by_none(self) -> list[GroupedNodes]:
        return [
            GroupedNodes(
                name=node.name,
                type="nodes",
                nodes=[node.name],
                dependencies=[
                    parent.name
                    for parent in sorted(
                        self.node_dependencies[node], key=lambda n: n.name
                    )
                ],
            )
            for node in self.nodes
        ]

    def _group_by_namespace(self) -> list[GroupedNodes]:
        grouped_nodes_map: dict[str, GroupedNodes] = {}

        for node in self.nodes:
            key = node.namespace.split(".")[0] if node.namespace else node.name

            if key not in grouped_nodes_map:
                grouped_nodes_map[key] = GroupedNodes(
                    name=key,
                    type="namespace" if node.namespace else "nodes",
                )

            grouped_nodes_map[key].nodes.append(node.name)

            dependencies = grouped_nodes_map[key].dependencies
            unique_dependencies = set(dependencies)

            for parent in sorted(self.node_dependencies[node], key=lambda n: n.name):
                parent_key = (
                    parent.namespace.split(".")[0] if parent.namespace else parent.name
                )
                if parent_key != key and parent_key not in unique_dependencies:
                    dependencies.append(parent_key)
                    unique_dependencies.add(parent_key)

        return list(grouped_nodes_map.values())

    def only_nodes(self, *node_names: str) -> Pipeline:
        """Create a new ``Pipeline`` which will contain only the specified
        nodes by name.

        Args:
            *node_names: One or more node names. The returned ``Pipeline``
                will only contain these nodes.

        Raises:
            ValueError: When some invalid node name is given.

        Returns:
            A new ``Pipeline``, containing only ``nodes``.

        """
        unregistered_nodes = set(node_names) - set(self._nodes_by_name.keys())
        if unregistered_nodes:
            # check if unregistered nodes are available under namespace
            namespaces = []
            for unregistered_node in unregistered_nodes:
                namespaces.extend(
                    [
                        node_name
                        for node_name in self._nodes_by_name.keys()
                        if node_name.endswith(f".{unregistered_node}")
                    ]
                )
            if namespaces:
                raise ValueError(
                    f"Pipeline does not contain nodes named {list(unregistered_nodes)}. "
                    f"Did you mean: {namespaces}?"
                )
            raise ValueError(
                f"Pipeline does not contain nodes named {list(unregistered_nodes)}."
            )

        nodes = [self._nodes_by_name[name] for name in node_names]
        return Pipeline(nodes)

    def only_nodes_with_namespaces(self, node_namespaces: list[str]) -> Pipeline:
        """Creates a new ``Pipeline`` containing only nodes with the specified
        namespaces.

        Args:
            node_namespaces: A list of node namespaces.

        Raises:
            ValueError: When pipeline contains no nodes with the specified namespaces.

        Returns:
            A new ``Pipeline`` containing nodes with the specified namespaces.
        """
        nodes = []
        unmatched_namespaces = []  # Track namespaces that don't match any nodes

        for node_namespace in node_namespaces:
            matching_nodes = []
            for n in self._nodes:
                if n.namespace and (
                    n.namespace == node_namespace
                    or n.namespace.startswith(f"{node_namespace}.")
                ):
                    matching_nodes.append(n)

            if not matching_nodes:
                unmatched_namespaces.append(node_namespace)
            nodes.extend(matching_nodes)

        if unmatched_namespaces:
            raise ValueError(
                f"Pipeline does not contain nodes with the following namespaces: {unmatched_namespaces}"
            )

        return Pipeline(nodes)

    def _get_nodes_with_inputs_transcode_compatible(
        self, datasets: set[str]
    ) -> set[Node]:
        """Retrieves nodes that use the given `datasets` as inputs.
        If provided a name, but no format, for a transcoded dataset, it
        includes all nodes that use inputs with that name, otherwise it
        matches to the fully-qualified name only (i.e. name@format).

        Raises:
            ValueError: if any of the given datasets do not exist in the
                ``Pipeline`` object

        Returns:
            Set of ``Nodes`` that use the given datasets as inputs.
        """
        missing = sorted(
            datasets - self.datasets() - self._transcode_compatible_names()
        )
        if missing:
            raise ValueError(f"Pipeline does not contain datasets named {missing}")

        relevant_nodes = set()
        for input_ in datasets:
            if _strip_transcoding(input_) == input_:
                relevant_nodes.update(self._nodes_by_input[_strip_transcoding(input_)])
            else:
                for node_ in self._nodes_by_input[_strip_transcoding(input_)]:
                    if input_ in node_.inputs:
                        relevant_nodes.add(node_)
        return relevant_nodes

    def _get_nodes_with_outputs_transcode_compatible(
        self, datasets: set[str]
    ) -> set[Node]:
        """Retrieves nodes that output to the given `datasets`.
        If provided a name, but no format, for a transcoded dataset, it
        includes the node that outputs to that name, otherwise it matches
        to the fully-qualified name only (i.e. name@format).

        Raises:
            ValueError: if any of the given datasets do not exist in the
                ``Pipeline`` object

        Returns:
            Set of ``Nodes`` that output to the given datasets.
        """
        missing = sorted(
            datasets - self.datasets() - self._transcode_compatible_names()
        )
        if missing:
            raise ValueError(f"Pipeline does not contain datasets named {missing}")

        relevant_nodes = set()
        for output in datasets:
            if _strip_transcoding(output) in self._nodes_by_output:
                node_with_output = self._nodes_by_output[_strip_transcoding(output)]
                if (
                    _strip_transcoding(output) == output
                    or output in node_with_output.outputs
                ):
                    relevant_nodes.add(node_with_output)

        return relevant_nodes

    def only_nodes_with_inputs(self, *inputs: str) -> Pipeline:
        """Create a new ``Pipeline`` object with the nodes which depend
        directly on the provided inputs.
        If provided a name, but no format, for a transcoded input, it
        includes all the nodes that use inputs with that name, otherwise it
        matches to the fully-qualified name only (i.e. name@format).

        Args:
            *inputs: A list of inputs which should be used as a starting
                point of the new ``Pipeline``.

        Raises:
            ValueError: Raised when any of the given inputs do not exist in the
                ``Pipeline`` object.

        Returns:
            A new ``Pipeline`` object, containing a subset of the
                nodes of the current one such that only nodes depending
                directly on the provided inputs are being copied.

        """
        starting = set(inputs)
        nodes = self._get_nodes_with_inputs_transcode_compatible(starting)

        return Pipeline(nodes)

    def from_inputs(self, *inputs: str) -> Pipeline:
        """Create a new ``Pipeline`` object with the nodes which depend
        directly or transitively on the provided inputs.
        If provided a name, but no format, for a transcoded input, it
        includes all the nodes that use inputs with that name, otherwise it
        matches to the fully-qualified name only (i.e. name@format).

        Args:
            *inputs: A list of inputs which should be used as a starting point
                of the new ``Pipeline``

        Raises:
            ValueError: Raised when any of the given inputs do not exist in the
                ``Pipeline`` object.

        Returns:
            A new ``Pipeline`` object, containing a subset of the
                nodes of the current one such that only nodes depending
                directly or transitively on the provided inputs are being
                copied.

        """
        starting = set(inputs)
        result: set[Node] = set()
        next_nodes = self._get_nodes_with_inputs_transcode_compatible(starting)

        while next_nodes:
            result |= next_nodes
            outputs = set(chain.from_iterable(node.outputs for node in next_nodes))
            starting = outputs

            next_nodes = set(
                chain.from_iterable(
                    self._nodes_by_input[_strip_transcoding(input_)]
                    for input_ in starting
                )
            )

        return Pipeline(result)

    def only_nodes_with_outputs(self, *outputs: str) -> Pipeline:
        """Create a new ``Pipeline`` object with the nodes which are directly
        required to produce the provided outputs.
        If provided a name, but no format, for a transcoded dataset, it
        includes all the nodes that output to that name, otherwise it matches
        to the fully-qualified name only (i.e. name@format).

        Args:
            *outputs: A list of outputs which should be the final outputs
                of the new ``Pipeline``.

        Raises:
            ValueError: Raised when any of the given outputs do not exist in the
                ``Pipeline`` object.

        Returns:
            A new ``Pipeline`` object, containing a subset of the nodes of the
            current one such that only nodes which are directly required to
            produce the provided outputs are being copied.
        """
        starting = set(outputs)
        nodes = self._get_nodes_with_outputs_transcode_compatible(starting)

        return Pipeline(nodes)

    def to_outputs(self, *outputs: str) -> Pipeline:
        """Create a new ``Pipeline`` object with the nodes which are directly
        or transitively required to produce the provided outputs.
        If provided a name, but no format, for a transcoded dataset, it
        includes all the nodes that output to that name, otherwise it matches
        to the fully-qualified name only (i.e. name@format).

        Args:
            *outputs: A list of outputs which should be the final outputs of
                the new ``Pipeline``.

        Raises:
            ValueError: Raised when any of the given outputs do not exist in the
                ``Pipeline`` object.


        Returns:
            A new ``Pipeline`` object, containing a subset of the nodes of the
            current one such that only nodes which are directly or transitively
            required to produce the provided outputs are being copied.

        """
        starting = set(outputs)
        result: set[Node] = set()
        next_nodes = self._get_nodes_with_outputs_transcode_compatible(starting)

        while next_nodes:
            result |= next_nodes
            inputs = set(chain.from_iterable(node.inputs for node in next_nodes))
            starting = inputs

            next_nodes = {
                self._nodes_by_output[_strip_transcoding(output)]
                for output in starting
                if _strip_transcoding(output) in self._nodes_by_output
            }

        return Pipeline(result)

    def from_nodes(self, *node_names: str) -> Pipeline:
        """Create a new ``Pipeline`` object with the nodes which depend
        directly or transitively on the provided nodes.

        Args:
            *node_names: A list of node_names which should be used as a
                starting point of the new ``Pipeline``.
        Raises:
            ValueError: Raised when any of the given names do not exist in the
                ``Pipeline`` object.
        Returns:
            A new ``Pipeline`` object, containing a subset of the nodes of
                the current one such that only nodes depending directly or
                transitively on the provided nodes are being copied.

        """

        res = self.only_nodes(*node_names)
        res += self.from_inputs(*map(_strip_transcoding, res.all_outputs()))
        return res

    def to_nodes(self, *node_names: str) -> Pipeline:
        """Create a new ``Pipeline`` object with the nodes required directly
        or transitively by the provided nodes.

        Args:
            *node_names: A list of node_names which should be used as an
                end point of the new ``Pipeline``.
        Raises:
            ValueError: Raised when any of the given names do not exist in the
                ``Pipeline`` object.
        Returns:
            A new ``Pipeline`` object, containing a subset of the nodes of the
                current one such that only nodes required directly or
                transitively by the provided nodes are being copied.

        """

        res = self.only_nodes(*node_names)
        res += self.to_outputs(*map(_strip_transcoding, res.all_inputs()))
        return res

    def only_nodes_with_tags(self, *tags: str) -> Pipeline:
        """Creates a new ``Pipeline`` object with the nodes which contain *any*
        of the provided tags. The resulting ``Pipeline`` is empty if no tags
        are provided.

        Args:
            *tags: A list of node tags which should be used to lookup
                the nodes of the new ``Pipeline``.
        Returns:
            Pipeline: A new ``Pipeline`` object, containing a subset of the
                nodes of the current one such that only nodes containing *any*
                of the tags provided are being copied.
        """
        unique_tags = set(tags)
        nodes = [node for node in self._nodes if unique_tags & node.tags]
        return Pipeline(nodes)

    def filter(  # noqa: PLR0913
        self,
        tags: Iterable[str] | None = None,
        from_nodes: Iterable[str] | None = None,
        to_nodes: Iterable[str] | None = None,
        node_names: Iterable[str] | None = None,
        from_inputs: Iterable[str] | None = None,
        to_outputs: Iterable[str] | None = None,
        node_namespaces: Iterable[str] | None = None,
    ) -> Pipeline:
        """Creates a new ``Pipeline`` object with the nodes that meet all of the
        specified filtering conditions.

        The new pipeline object is the intersection of pipelines that meet each
        filtering condition. This is distinct from chaining multiple filters together.

        Args:
            tags: A list of node tags which should be used to lookup
                the nodes of the new ``Pipeline``.
            from_nodes: A list of node names which should be used as a
                starting point of the new ``Pipeline``.
            to_nodes:  A list of node names which should be used as an
                end point of the new ``Pipeline``.
            node_names: A list of node names which should be selected for the
                new ``Pipeline``.
            from_inputs: A list of inputs which should be used as a starting point
                of the new ``Pipeline``
            to_outputs: A list of outputs which should be the final outputs of
                the new ``Pipeline``.
            node_namespaces: A list of node namespaces which should be used to select
                nodes in the new ``Pipeline``.

        Returns:
            A new ``Pipeline`` object with nodes that meet all of the specified
                filtering conditions.

        Raises:
            ValueError: The filtered ``Pipeline`` has no nodes.

        Example:
        ``` python
        pipeline = Pipeline(
            [
                node(func, "A", "B", name="node1"),
                node(func, "B", "C", name="node2"),
                node(func, "C", "D", name="node3"),
            ]
        )
        pipeline.filter(node_names=["node1", "node3"], from_inputs=["A"])
        # Gives a new pipeline object containing node1 and node3.
        ```
        """

        filter_methods = {
            self.only_nodes_with_tags: tags,
            self.from_nodes: from_nodes,
            self.to_nodes: to_nodes,
            self.only_nodes: node_names,
            self.from_inputs: from_inputs,
            self.to_outputs: to_outputs,
            self.only_nodes_with_namespaces: [node_namespaces]
            if node_namespaces
            else None,
        }

        subset_pipelines = {
            filter_method(*filter_args)  # type: ignore
            for filter_method, filter_args in filter_methods.items()
            if filter_args
        }

        # Intersect all the pipelines subsets. We apply each filter to the original
        # pipeline object (self) rather than incrementally chaining filter methods
        # together. Hence, the order of filtering does not affect the outcome, and the
        # resultant pipeline is unambiguously defined.
        # If this were not the case then, for example,
        # pipeline.filter(node_names=["node1", "node3"], from_inputs=["A"])
        # would give different outcomes depending on the order of filter methods:
        # only_nodes and then from_inputs would give node1, while only_nodes and then
        # from_inputs would give node1 and node3.
        filtered_pipeline = Pipeline(self._nodes)
        for subset_pipeline in subset_pipelines:
            filtered_pipeline &= subset_pipeline

        if not filtered_pipeline.nodes:
            raise ValueError(
                "Pipeline contains no nodes after applying all provided filters. "
                "Please ensure that at least one pipeline with nodes has been defined."
            )
        return filtered_pipeline

    def tag(self, tags: str | Iterable[str]) -> Pipeline:
        """Tags all the nodes in the pipeline.

        Args:
            tags: The tags to be added to the nodes.

        Returns:
            New ``Pipeline`` object with nodes tagged.
        """
        nodes = [n.tag(tags) for n in self._nodes]
        return Pipeline(nodes)

    def to_json(self) -> str:
        """Return a json representation of the pipeline."""
        transformed = [
            {
                "name": n.name,
                "inputs": list(n.inputs),
                "outputs": list(n.outputs),
                "tags": list(n.tags),
            }
            for n in self._nodes
        ]
        pipeline_versioned = {
            "kedro_version": kedro.__version__,
            "pipeline": transformed,
        }

        return json.dumps(pipeline_versioned)

    def _rename(
        self,
        name: str,
        mapping: dict,
        namespace: str | None,
        prefix_datasets_with_namespace: bool,
    ) -> str:
        def _prefix_dataset(name: str) -> str:
            return f"{namespace}.{name}"

        def _prefix_param(name: str) -> str:
            _, param_name = name.split("params:")
            return f"params:{namespace}.{param_name}"

        def _is_transcode_base_in_mapping(name: str) -> bool:
            base_name, _ = _transcode_split(name)
            return base_name in mapping

        def _map_transcode_base(name: str) -> str:
            from kedro.pipeline.transcoding import TRANSCODING_SEPARATOR

            base_name, transcode_suffix = _transcode_split(name)
            return TRANSCODING_SEPARATOR.join((mapping[base_name], transcode_suffix))

        rules = [
            # if name mapped to new name, update with new name
            (lambda n: n in mapping, lambda n: mapping[n]),
            # if name refers to the set of all "parameters", leave as is
            (_is_all_parameters, lambda n: n),
            # if transcode base is mapped to a new name, update with new base
            (_is_transcode_base_in_mapping, _map_transcode_base),
        ]

        # Add rules for prefixing only if prefix_datasets_with_namespace is True
        if prefix_datasets_with_namespace:
            rules.extend(
                [
                    # if name refers to a single parameter and a namespace is given, apply prefix
                    (
                        lambda n: bool(namespace) and _is_single_parameter(n),
                        _prefix_param,
                    ),
                    # if namespace given for a dataset, prefix name using that namespace
                    (lambda n: bool(namespace), _prefix_dataset),
                ]
            )

        for predicate, processor in rules:
            if predicate(name):  # type: ignore[no-untyped-call]
                processor_name: str = processor(name)  # type: ignore[no-untyped-call]
                return processor_name

        # leave name as is
        return name

    def _process_dataset_names(
        self,
        datasets: str | list[str] | dict[str, str] | None,
        mapping: dict,
        namespace: str | None,
        prefix_datasets_with_namespace: bool = True,
    ) -> str | list[str] | dict[str, str] | None:
        if datasets is None:
            return None
        if isinstance(datasets, str):
            return self._rename(
                datasets, mapping, namespace, prefix_datasets_with_namespace
            )
        if isinstance(datasets, list):
            return [
                self._rename(name, mapping, namespace, prefix_datasets_with_namespace)
                for name in datasets
            ]
        if isinstance(datasets, dict):
            return {
                key: self._rename(
                    value, mapping, namespace, prefix_datasets_with_namespace
                )
                for key, value in datasets.items()
            }
        raise ValueError(
            f"Unexpected input {datasets} of type {type(datasets)}"
        )  # pragma: no cover

    def _copy_node(
        self,
        node: Node,
        mapping: dict,
        namespace: str | None,
        prefix_datasets_with_namespace: bool = True,
    ) -> Node:
        new_namespace = node.namespace
        if namespace:
            new_namespace = (
                f"{namespace}.{node.namespace}" if node.namespace else namespace
            )
        return node._copy(
            inputs=self._process_dataset_names(
                node._inputs, mapping, namespace, prefix_datasets_with_namespace
            ),
            outputs=self._process_dataset_names(
                node._outputs, mapping, namespace, prefix_datasets_with_namespace
            ),
            namespace=new_namespace,
            confirms=self._process_dataset_names(
                node._confirms, mapping, namespace, prefix_datasets_with_namespace
            ),
        )

    def _map_nodes(  # noqa: PLR0913
        self,
        nodes: Iterable[Node | Pipeline] | Pipeline,
        inputs: str | set[str] | dict[str, str] | None = None,
        outputs: str | set[str] | dict[str, str] | None = None,
        parameters: str | set[str] | dict[str, str] | None = None,
        tags: str | Iterable[str] | None = None,
        namespace: str | None = None,
        prefix_datasets_with_namespace: bool = True,
    ) -> list[Node]:
        """Map namespace to the inputs, outputs, parameters and nodes of the pipeline."""
        if isinstance(nodes, Pipeline):
            # To ensure that we are always dealing with a *copy* of pipe.
            pipe = Pipeline([nodes], tags=tags)
        else:
            pipe = Pipeline(nodes, tags=tags)

        inputs = _get_dataset_names_mapping(inputs)
        outputs = _get_dataset_names_mapping(outputs)
        parameters = _get_param_names_mapping(parameters)

        _validate_datasets_exist(inputs.keys(), outputs.keys(), parameters.keys(), pipe)
        _validate_inputs_outputs(inputs.keys(), outputs.keys(), pipe)

        mapping = {**inputs, **outputs, **parameters}
        new_nodes = [
            self._copy_node(n, mapping, namespace, prefix_datasets_with_namespace)
            for n in pipe.nodes
        ]
        return new_nodes


def pipeline(  # noqa: PLR0913
    nodes: Iterable[Node | Pipeline] | Pipeline,
    *,
    inputs: str | set[str] | dict[str, str] | None = None,
    outputs: str | set[str] | dict[str, str] | None = None,
    parameters: str | set[str] | dict[str, str] | None = None,
    tags: str | Iterable[str] | None = None,
    namespace: str | None = None,
    prefix_datasets_with_namespace: bool = True,
) -> Pipeline:
    r"""Create a ``Pipeline`` from a collection of nodes and/or ``Pipeline``\s.

    Args:
        nodes: The nodes the ``Pipeline`` will be made of. If you
            provide pipelines among the list of nodes, those pipelines will
            be expanded and all their nodes will become part of this
            new pipeline.
        inputs: A name or collection of input names to be exposed as connection points
            to other pipelines upstream. This is optional; if not provided, the
            pipeline inputs are automatically inferred from the pipeline structure.
            When str or set[str] is provided, the listed input names will stay
            the same as they are named in the provided pipeline.
            When dict[str, str] is provided, current input names will be
            mapped to new names.
            Must only refer to the pipeline's free inputs.
        outputs: A name or collection of names to be exposed as connection points
            to other pipelines downstream. This is optional; if not provided, the
            pipeline outputs are automatically inferred from the pipeline structure.
            When str or set[str] is provided, the listed output names will stay
            the same as they are named in the provided pipeline.
            When dict[str, str] is provided, current output names will be
            mapped to new names.
            Can refer to both the pipeline's free outputs, as well as
            intermediate results that need to be exposed.
        parameters: A name or collection of parameters to namespace.
            When str or set[str] are provided, the listed parameter names will stay
            the same as they are named in the provided pipeline.
            When dict[str, str] is provided, current parameter names will be
            mapped to new names.
            The parameters can be specified without the `params:` prefix.
        tags: Optional set of tags to be applied to all the pipeline nodes.
        namespace: A prefix to give to all dataset names,
            except those explicitly named with the `inputs`/`outputs`
            arguments, and parameter references (`params:` and `parameters`).
        prefix_datasets_with_namespace: A flag to specify if the inputs and outputs of the nodes
                should be prefixed with the namespace. It is set to True by default. It is
                useful to turn off when namespacing is used for grouping nodes for deployment purposes.

    Raises:
        PipelineError: When inputs, outputs or parameters are incorrectly
            specified, or they do not exist on the original pipeline.
        ValueError: When underlying pipeline nodes inputs/outputs are not
            any of the expected types (str, dict, list, or None).

    Returns:
        A new ``Pipeline`` object.
    """

    return Pipeline(
        nodes=nodes,
        inputs=inputs,
        outputs=outputs,
        parameters=parameters,
        tags=tags,
        namespace=namespace,
        prefix_datasets_with_namespace=prefix_datasets_with_namespace,
    )


def _get_dataset_names_mapping(
    names: str | set[str] | dict[str, str] | None = None,
) -> dict[str, str]:
    """Take a name or a collection of dataset names
    and turn it into a mapping from the old dataset names to the provided ones if necessary.

    Args:
        names: A dataset name or collection of dataset names.
            When str or set[str] is provided, the listed names will stay
            the same as they are named in the provided pipeline.
            When dict[str, str] is provided, current names will be
            mapped to new names in the resultant pipeline.

    Returns:
        A dictionary that maps the old dataset names to the provided ones.

    Examples:
    ``` python
    _get_dataset_names_mapping("dataset_name")
    {"dataset_name": "dataset_name"}  # a str name will stay the same
    _get_dataset_names_mapping(set(["ds_1", "ds_2"]))
    {"ds_1": "ds_1", "ds_2": "ds_2"}  # a set[str] of names will stay the same
    _get_dataset_names_mapping({"ds_1": "new_ds_1_name"})
    {"ds_1": "new_ds_1_name"}  # a dict[str, str] of names will map key to value
    ```
    """
    if names is None:
        return {}
    if isinstance(names, str):
        return {names: names}
    if isinstance(names, dict):
        return copy.deepcopy(names)

    return {item: item for item in names}


def _normalize_param_name(name: str) -> str:
    """Make sure that a param name has a `params:` prefix before passing to the node"""
    return name if name.startswith("params:") else f"params:{name}"


def _get_param_names_mapping(
    names: str | set[str] | dict[str, str] | None = None,
) -> dict[str, str]:
    """Take a parameter or a collection of parameter names
    and turn it into a mapping from existing parameter names to new ones if necessary.
    It follows the same rule as `_get_dataset_names_mapping` and
    prefixes the keys on the resultant dictionary with `params:` to comply with node's syntax.

    Args:
        names: A parameter name or collection of parameter names.
            When str or set[str] is provided, the listed names will stay
            the same as they are named in the provided pipeline.
            When dict[str, str] is provided, current names will be
            mapped to new names in the resultant pipeline.

    Returns:
        A dictionary that maps the old parameter names to the provided ones.

    Examples:
    ``` python
    _get_param_names_mapping("param_name")
    {"params:param_name": "params:param_name"}  # a str name will stay the same
    _get_param_names_mapping(set(["param_1", "param_2"]))
    # a set[str] of names will stay the same
    {"params:param_1": "params:param_1", "params:param_2": "params:param_2"}
    _get_param_names_mapping({"param_1": "new_name_for_param_1"})
    # a dict[str, str] of names will map key to value
    {"params:param_1": "params:new_name_for_param_1"}
    ```
    """
    params = {}
    for name, new_name in _get_dataset_names_mapping(names).items():
        if _is_all_parameters(name):
            params[name] = name  # don't map parameters into params:parameters
        else:
            param_name = _normalize_param_name(name)
            param_new_name = _normalize_param_name(new_name)
            params[param_name] = param_new_name
    return params


def _validate_duplicate_nodes(nodes_or_pipes: Iterable[Node | Pipeline]) -> None:
    seen_nodes: set[str] = set()
    duplicates: dict[Pipeline | None, set[str]] = defaultdict(set)

    def _check_node(node_: Node, pipeline_: Pipeline | None = None) -> None:
        name = node_.name
        if name in seen_nodes:
            duplicates[pipeline_].add(name)
        else:
            seen_nodes.add(name)

    for each in nodes_or_pipes:
        if isinstance(each, Node):
            _check_node(each)
        elif isinstance(each, Pipeline):
            for node in each.nodes:
                _check_node(node, pipeline_=each)

    if duplicates:
        duplicates_info = ""

        for pipeline, names in duplicates.items():
            pipe_repr = (
                "Free nodes" if pipeline is None else repr(pipeline).replace("\n", "")
            )
            nodes_repr = "\n".join(f"  - {name}" for name in sorted(names))
            duplicates_info += f"{pipe_repr}:\n{nodes_repr}\n"

        raise ValueError(
            f"Pipeline nodes must have unique names. The following node names "
            f"appear more than once:\n\n{duplicates_info}\nYou can name your "
            f"nodes using the last argument of 'node()'."
        )


def _validate_unique_outputs(nodes: list[Node]) -> None:
    outputs_chain = chain.from_iterable(node.outputs for node in nodes)
    outputs = map(_strip_transcoding, outputs_chain)
    duplicates = [key for key, value in Counter(outputs).items() if value > 1]
    if duplicates:
        raise OutputNotUniqueError(
            f"Output(s) {sorted(duplicates)} are returned by more than one nodes. Node "
            f"outputs must be unique."
        )


def _validate_unique_confirms(nodes: list[Node]) -> None:
    confirms_chain = chain.from_iterable(node.confirms for node in nodes)
    confirms = map(_strip_transcoding, confirms_chain)
    duplicates = [key for key, value in Counter(confirms).items() if value > 1]
    if duplicates:
        raise ConfirmNotUniqueError(
            f"{sorted(duplicates)} datasets are confirmed by more than one node. Node "
            f"confirms must be unique."
        )


def _validate_transcoded_inputs_outputs(nodes: list[Node]) -> None:
    """Users should not be allowed to refer to a transcoded dataset both
    with and without the separator.
    """
    all_inputs_outputs = set(
        chain(
            chain.from_iterable(node.inputs for node in nodes),
            chain.from_iterable(node.outputs for node in nodes),
        )
    )

    invalid = set()
    for dataset_name in all_inputs_outputs:
        name = _strip_transcoding(dataset_name)
        if name != dataset_name and name in all_inputs_outputs:
            invalid.add(name)

    if invalid:
        raise ValueError(
            f"The following datasets are used with transcoding, but "
            f"were referenced without the separator: {', '.join(invalid)}.\n"
            f"Please specify a transcoding option or "
            f"rename the datasets."
        )


def _match_namespaces(node_namespace: str, filter_namespace: str) -> bool:
    return node_namespace.split(".")[
        : len(filter_namespace.split("."))
    ] == filter_namespace.split(".")


class CircularDependencyError(Exception):
    """Raised when it is not possible to provide a topological execution
    order for nodes, due to a circular dependency existing in the node
    definition.
    """

    pass



import copy
import logging
import sys
from logging import LogRecord
from pathlib import Path
from typing import Any

import click
import rich.logging
import rich.pretty
import rich.traceback

from kedro.utils import _format_rich, _is_databricks


class RichHandler(rich.logging.RichHandler):
    """Identical to rich's logging handler but with a few extra behaviours:
    * warnings issued by the `warnings` module are redirected to logging
    * pretty printing is enabled on the Python REPL (including IPython and Jupyter)
    * all tracebacks are handled by rich when rich_tracebacks=True
    * constructor's arguments are mapped and passed to `rich.traceback.install`

    The list of available options of ``RichHandler`` can be found here:
    https://rich.readthedocs.io/en/stable/reference/logging.html#rich.logging.RichHandler

    The list of available options of `rich.traceback.install` can be found here:
    https://rich.readthedocs.io/en/stable/reference/traceback.html#rich.traceback.install
    """

    def __init__(self, *args: Any, **kwargs: Any):
        if "markup" not in kwargs:
            kwargs.update({"markup": True})  # Set markup as default
        super().__init__(*args, **kwargs)
        logging.captureWarnings(True)
        rich.pretty.install()

        # We suppress click here to hide tracebacks related to it conversely,
        # kedro is not suppressed to show its tracebacks for easier debugging.
        # sys.executable is used to get the kedro executable path to hide the
        # top level traceback.

        traceback_install_kwargs = {
            "suppress": [click, str(Path(sys.executable).parent)]
        }

        # Mapping arguments from RichHandler's Constructor to rich.traceback.install
        prefix = "tracebacks_"
        for key, value in kwargs.items():
            if key.startswith(prefix):
                key_prefix_removed = key[len(prefix) :]
                if key_prefix_removed == "suppress":
                    traceback_install_kwargs[key_prefix_removed].extend(value)
                else:
                    traceback_install_kwargs[key_prefix_removed] = value

        if self.rich_tracebacks and not _is_databricks():
            # Rich traceback handling does not work on databricks. Hopefully this will be
            # fixed on their side at some point, but until then we disable it.
            # See https://github.com/Textualize/rich/issues/2455
            rich.traceback.install(**traceback_install_kwargs)  # type: ignore[arg-type]

    def emit(self, record: LogRecord) -> None:
        args = record.args
        if not args or not hasattr(record, "rich_format"):
            return super().emit(record)

        else:
            # LogRecord is shared among all handler, created a new instance so it does not affect the rest.
            updated_record = copy.copy(record)
            format_args = record.rich_format  # type: ignore

            if format_args and isinstance(format_args, list):
                new_args = list(args)
                # Add markup to the message
                # if the list is shorter than the arg list, keep it unchanged
                for i, color in enumerate(format_args):
                    new_args[i] = _format_rich(str(new_args[i]), color)
                updated_record.args = tuple(new_args)
                super().emit(updated_record)
            else:
                raise TypeError("rich_format only accept non-empty list as an argument")



"""This module provides ``kedro.abstract_config`` with the baseline
class model for a `ConfigLoader` implementation.
"""

from __future__ import annotations

from collections import UserDict
from pathlib import Path  # noqa: TCH003
from typing import Any


class AbstractConfigLoader(UserDict):
    """``AbstractConfigLoader`` is the abstract base class
        for all `ConfigLoader` implementations.
    All user-defined `ConfigLoader` implementations should inherit
        from `AbstractConfigLoader` and implement all relevant abstract methods.
    """

    def __init__(
        self,
        conf_source: str | Path,
        env: str | None = None,
        runtime_params: dict[str, Any] | None = None,
        **kwargs: Any,
    ):
        super().__init__()
        self.conf_source = conf_source
        self.env = env
        self.runtime_params = runtime_params or {}

    # As of Python 3.12 __getitem__ is no longer called in the inherited UserDict.get()
    # This causes AbstractConfigLoader.get() to break
    # See: https://github.com/python/cpython/issues/105524
    # Overwrite the inherited get function with the implementation from 3.11 and prior
    def get(self, key: str, default: Any = None) -> Any:
        "D.get(k[,d]) -> D[k] if k in D, else d.  d defaults to None."
        try:
            return self[key]
        except KeyError:
            return default


class BadConfigException(Exception):
    """Raised when a configuration file cannot be loaded, for instance
    due to wrong syntax or poor formatting.
    """

    pass


class MissingConfigException(Exception):
    """Raised when no configuration files can be found within a config path"""

    pass
